# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
linear optical effects

"""
from __future__ import print_function
from __future__ import absolute_import

import warnings
# import math

import numpy as np
import copy
import numba

import time


#==============================================================================
# GLOBAL PARAMETERS
#==============================================================================




#==============================================================================
# EXCEPTIONS
#==============================================================================






# =============================================================================
# general field propagation routines
# =============================================================================
@numba.njit(parallel=True, cache=True)
def _calc_repropagation(P, Escat, G_dyad_list):
    if len(P) != G_dyad_list.shape[2]//3:
        raise Exception("polarization and Greens tensor arrays don't match in size!")
    _P = P.flatten()
    for i_p_r in numba.prange(G_dyad_list.shape[0]):
        Escat[i_p_r] = np.dot(G_dyad_list[i_p_r], _P)
    
    





# =============================================================================
# farfield functions
# =============================================================================
def _calc_extinct(wavelength, n_env, alpha_tensor, E0, E, 
                  with_radiation_correction=False, normalization_E0=True):
    """actual calculation of extinction and absorption cross sections
    
    numba compatible. tested.
    """
    if normalization_E0:
        I0_norm = np.sum(np.abs(E0)**2, axis=1).max()
    else:
        I0_norm = 1
    
    P = np.matmul(alpha_tensor, E[...,None])[...,0]
    
    cext = (1/I0_norm) * ((8 * np.pi**2 / wavelength) / n_env * 
            np.sum(np.multiply(np.conjugate(E0), P))).imag
    cabs = (1/I0_norm) * ((8 * np.pi**2 / wavelength) / n_env * 
            (np.sum(np.multiply(P, np.conjugate(E)))).imag).real
    
    if with_radiation_correction:
        ak0 = 2*np.pi / wavelength
        cabs -= ((8 * np.pi**2 / wavelength) / n_env * 
                 (2/3)*ak0**3 * np.sum(np.multiply(P, np.conjugate(P))))
    csca = cext - cabs
    
    return cext, csca, cabs
    

def extinct(sim, field_index, with_radiation_correction=False, normalization_E0=True):
    """Extinction, scattering and absorption cross-sections
    
    Calculates extinction, scattering and absorption crosssections
    for each wavelength of the GDM simulation
    
    Pure python implementation.
    
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    with_radiation_correction : bool, default: False
        Adds an optional radiative correction to the absorption section (hence 
        it interferes also in the scattering section).See 
        *B. Draine, in: Astrophys. J. 333, 848–872 (1988)*, equation (3.06).
        Using the correction can lead to better agreement for very large 
        discretization stepsizes, but has only a weak influence on simulations
        with fine meshes.
        
    normalization_E0 : bool, default: True
        normalizes sections to peak of incident field intensity inside structure
    
    Returns
    -------
    ext-section : float
        extinction cross-section
    
    scat-section : float
        scattering cross-section
    
    abs-section : float
        apsorption cross-section
        
    
    Notes
    -----
    For the calculation of the cross-sections from the complex nearfield, 
    see e.g.: 
    Draine, B. T. & Flatau, P. J. **Discrete-dipole approximation for scattering 
    calculations.**
    Journal of the Optical Society of America, A 11, 1491 (1994).
    
    """
    from pyGDM2 import tools 
    if sim.E is None: 
        raise ValueError("Error: Scattered field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
    
    ## --- incident field configuration
    field_params = tools.get_field_indices(sim)[field_index]
    wavelength   = field_params['wavelength']
    sim.struct.setWavelength(wavelength)
    eps_env = sim.dyads.getEnvironmentIndices(wavelength, sim.struct.geometry[:1])[0]  # assume structure is fully in one environment
    n_env = np.sqrt(eps_env)
    
    ## --- get polarizability at wavelength
    alpha_tensor = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)
    
    env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    E0 = sim.efield.field_generator(sim.struct.geometry, 
                                    env_dict, **field_params)
    E = sim.E[field_index][1]
    
    return _calc_extinct(wavelength, n_env, alpha_tensor, E0, E, 
                         with_radiation_correction, normalization_E0)
    


def farfield(sim, field_index, 
                r_probe=None,
                r=100000., 
                tetamin=0, tetamax=np.pi/2., Nteta=10, 
                phimin=0, phimax=2*np.pi, Nphi=36, 
                polarizerangle='none', return_value='map', 
                normalization_E0=False):
    """spatially resolved and polarization-filtered far-field scattering 
    
    For a given incident field, calculate the electro-magnetic field 
    (E-component) in the far-field around the nanostructure 
    (on a sphere of radius `r`).
    
    Propagator for scattering into a substrate contributed by C. Majorel
    
    Pure python implementation.
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
        
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    r_probe : tuple (x,y,z) or list of 3-lists/-tuples. optional. Default: don't use
        defaults to *None*, which means it is not used and a solid angle defined by
        a spherical coordinate range is used instead. If `r_probe` is given, this
        overrides `r`, `tetamin`, `tetamax`, `Nteta`, `Nphi`.
        (list of) coordinate(s) to evaluate farfield on. 
        Format: tuple (x,y,z) or list of 3 lists: [Xmap, Ymap, Zmap] 
        
    r : float, default: 100000.
        radius of integration sphere (distance to coordinate origin in nm)
        
    tetamin, tetamax : float, float; defaults: 0, np.pi/2
        minimum and maximum polar angle in radians 
        (in linear steps from `tetamin` to `tetamax`)
        
    phimin, phimax : float, float; defaults: 0, 2*np.pi
        minimum and maximum azimuth angle in radians, excluding last position
        (in linear steps from `phimin` to `phimax`)
        
    Nteta, Nphi : int, int; defaults: 10, 36
        number of polar and azimuthal angles on sphere to calculate,
        
    polarizerangle : float or 'none', default: 'none'
        optional polarization filter angle **in degrees**(!). If 'none' (default), 
        the total field-intensity is calculated (= no polarization filter)
    
    return_value : str, default: 'map'
        Values to be returned. Either 'map' (default) or 'integrated'.
          - "map" : (default) return spatially resolved farfield intensity at each spherical coordinate (5 lists)
          - "efield" : return spatially resolved E-fields at each spherical coordinate (5 lists)
          - "int_Es" : return the integrated scattered field (as float)
          - "int_E0" : return the integrated fundamental field (as float)
          - "int_Etot" : return the integrated total field (as float)
          
    normalization_E0 : bool, default: False
        has only effect on return_value=="int_Es": Normalizes scattering to peak of incident field intensity inside structure
    
    
    Returns
    -------
    using `r_probe` for position definition:
        3 lists of 6-tuples (x,y,z, Ex,Ey,Ez), complex : 
            - scattered Efield or E-intensity
            - total Efield (inlcuding fundamental field) or total E intensity
            - fundamental Efield (incident field) or E0 intensity
        
    if solid angle is defined via spherical coordinate range:
        - return_value == "map" : 5 arrays of shape (Nteta, Nphi) : 
            - [tetalist : teta angles - if solid angle range input]
            - [philist : phi angles - if solid angle range input]
            - I_sc : intensity of scattered field, 
            - I_tot : intensity of total field (I_tot=|E_sc+E_0|^2), 
            - I0 : intensity of incident field
        
        - return_value == "efield" : float
            - [tetalist : teta angles - if solid angle range input]
            - [philist : phi angles - if solid angle range input]
            - E_sc : complex scattered field at each pos.
            - E_tot : complex total field at each pos. (E_sc+E0)
            - E0 : complex incident field at each pos.
            
        - return_value == "int_XX" : float
            integrated total intensity over specified solid angle
        
    Notes
    -----
    For details of the asymptotic (non-retarded) far-field propagators for a 
    dipole above a substrate, see e.g.:
    Colas des Francs, G. & Girard, C. & Dereux, A. **Theory of near-field 
    optical imaging with a single molecule as light source.** 
    The Journal of Chemical Physics 117, 4659–4666 (2002)
        
    """
# =============================================================================
#     exception handling
# =============================================================================
    if sim.E is None: 
        raise ValueError("Error: Scattering field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
    
    if str(polarizerangle).lower() == 'none':
        polarizer = 0
    else:
        polarizer = polarizerangle * np.pi/180.
    
    if r_probe is not None and return_value in ['int_es', 'int_E0', 'int_Etot']:
        raise ValueError("probing farfield on user-defined positions does not support integration " +
                         "of the intensity since no surface differential can be defined. Use spherical " +
                         "coordinate definition to do the integration.")
    # --- check angles
    if return_value.lower().startswith('int') and (0 > tetamin >= tetamax):
        raise ValueError("`tetamin` out of range, must be in [0, tetamax] for integration.")
    
    if return_value.lower().startswith('int') and (np.pi < tetamax < 0):
        raise ValueError("`tetamax` out of range, must be in [0, pi] for integration.")
    
    if phimin < 0:
        warnings.warn("phi is supposed to be a positive number, use at own risk.")
    if phimin > phimax:
        raise ValueError("`phimin` out of range, must be <= phimax")
    
    if return_value.lower().startswith('int') and (0 > phimax > 2*np.pi):
        raise ValueError("`phimax` out of range, must be in [0, 2*pi] for integration.")
    
    
    
# =============================================================================
#     preparation
# =============================================================================
    from pyGDM2 import tools

    ## --- spherical probe coordinates
    if r_probe is None:
        tetalist = np.ones((int(Nteta), int(Nphi)))*np.linspace(tetamin, tetamax, int(Nteta))[:,None]
        philist = np.ones((int(Nteta), int(Nphi)))*np.linspace(phimin, phimax, int(Nphi), endpoint=False)[None,:]
        xff = (r * np.sin(tetalist) * np.cos(philist)).flatten()
        yff = (r * np.sin(tetalist) * np.sin(philist)).flatten()
        zff = (r * np.cos(tetalist)).flatten()
        _r_probe = np.transpose([xff, yff, zff])
    else:
        _r_probe = r_probe
        
    ## --- spherical integration steps
    dteta = (tetamax-tetamin) / float(Nteta-1)
    dphi = (phimax-phimin) / float(Nphi)
    
    ## --- incident field config
    field_params    = tools.get_field_indices(sim)[field_index]
    wavelength      = field_params['wavelength']
    
    ## --- environment
    sim.struct.setWavelength(wavelength)
    conf_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    if sim.dyads.n2_material.__name__ != sim.dyads.n3_material.__name__:
        warnings.warn("`farfield` does not support a cladding layer at the moment. " +
                      "It implements only an asymptotic Green's tensor for an " +
                      "environment with a single interface (substrate). The results are probably incorrect.")
    
#==============================================================================
#     electric polarization of structure, fundamental field
#==============================================================================        
    ## --- fundamental field - use dummy structure with 
    env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    if return_value == 'int_Es':
        from pyGDM2 import fields
        E0 = fields.nullfield(_r_probe, env_dict, wavelength, returnField='E')
    else:
        E0 = sim.efield.field_generator(_r_probe, env_dict, **field_params)
        
    
    ## --- electric polarization of each discretization cell via tensorial polarizability
    Eint = sim.E[field_index][1]
    alpha_tensor = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)
    P = np.matmul(alpha_tensor, Eint[...,None])[...,0]
    
    
    ## --- Greens function for each dipole
    G_FF_EE = np.zeros((len(sim.struct.geometry), len(_r_probe), 3, 3), 
                       dtype = sim.efield.dtypec)
    sim.dyads.eval_G(sim.struct.geometry, _r_probe, 
                          sim.dyads.G_EE_ff, wavelength, 
                          conf_dict, G_FF_EE)
    
    ## propagate fields 
    G_FF_EE = np.moveaxis(G_FF_EE, 0,2).reshape((len(_r_probe), 3, -1))  # re-arrange for direct matrix-vector multiplication
    Escat = np.zeros(shape=(len(_r_probe), 3), dtype=sim.efield.dtypec)
    _calc_repropagation(P, Escat, G_FF_EE)
    
    Iscat = np.sum((np.abs(Escat)**2), axis=1)
    
    
    
#==============================================================================
#    calc. fields through optional polarization filter
#==============================================================================
    if str(polarizerangle).lower() != 'none':
        ## --- scattered E-field parallel and perpendicular to scattering plane
        Es_par  = ( Escat.T[0] * np.cos(tetalist.flatten()) * np.cos(philist.flatten()) + 
                    Escat.T[1] * np.sin(philist.flatten()) * np.cos(tetalist.flatten()) - 
                    Escat.T[2] * np.sin(tetalist.flatten()) )
        Es_perp = ( Escat.T[0] * np.sin(philist.flatten()) - Escat.T[1] * np.cos(philist.flatten()) )
        ## --- scattered E-field parallel to polarizer
        Es_pol  = ( Es_par * np.cos(polarizer - philist.flatten()) - 
                    Es_perp * np.sin(polarizer - philist.flatten()) )
        
        ## --- fundamental E-field parallel and perpendicular to scattering plane
        E0_par  = ( E0.T[0] * np.cos(tetalist.flatten()) * np.cos(philist.flatten()) + 
                    E0.T[1] * np.sin(philist.flatten()) * np.cos(tetalist.flatten()) - 
                    E0.T[2] * np.sin(tetalist.flatten()) )
        E0_perp = ( E0.T[0] * np.sin(philist.flatten()) - E0.T[1] * np.cos(philist.flatten()) )
        ## --- fundamental E-field parallel to polarizer
        E0_pol  = ( E0_par * np.cos(polarizer - philist.flatten()) - 
                    E0_perp * np.sin(polarizer - philist.flatten()) )

#==============================================================================
#     Intensities with and without fundamental field / polarizer
#==============================================================================
    ## --- total field (no polarizer)
    if r_probe is None:
        out_shape = tetalist.shape
    else:
        out_shape = len(_r_probe)
        
    I_sc  = Iscat.reshape(out_shape)
    I0    = np.sum((np.abs(E0)**2), axis=1).reshape(out_shape)
    I_tot = np.sum((np.abs(E0 + Escat)**2), axis=1).reshape(out_shape)
    
    ## --- optionally: with polarizer
    if str(polarizerangle).lower() != 'none':
        I_sc  = (np.abs(Es_pol)**2).reshape(out_shape)
        I0    = (np.abs(E0_pol)**2).reshape(out_shape)
        I_tot = (np.abs(Es_pol + E0_pol)**2).reshape(out_shape)
    
    
    if return_value.lower() == 'map':
        if r_probe is None:
            return tetalist, philist, I_sc, I_tot, I0
        else:
            return I_sc, I_tot, I0
    elif return_value.lower() in ['efield', 'fields', 'field']:
        if r_probe is None:
            return tetalist, philist, Escat, Escat + E0, E0
        else:
            return Escat, Escat + E0, E0
    else:
        d_solid_surf = r**2 * np.sin(tetalist) * dteta * dphi
        if return_value.lower() == 'int_es':
            if normalization_E0:
                env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
                E0 = sim.efield.field_generator(sim.struct.geometry, 
                                                env_dict, **field_params)
                I0_norm = np.sum(np.abs(E0)**2, axis=1).max()
            else:
                I0_norm = 1
            
            return np.sum(I_sc * d_solid_surf) / I0_norm
        
        elif return_value.lower() == 'int_e0':
            return np.sum(I0 * d_solid_surf)
        
        elif return_value.lower() == 'int_etot':
            return np.sum(I_tot * d_solid_surf)
        
        else:
            raise ValueError("Parameter 'return_value' must be one of ['map', 'int_es', 'int_e0', 'int_etot'].")





# =============================================================================
# nearfield functions
# =============================================================================
def nearfield(sim, field_index, r_probe, which_fields=["Es","Et","Bs","Bt"],
              val_inside_struct="field", val_outside_struct="field", 
              N_neighbors_internal_field=4,
              method=''):
    """Nearfield distribution in proximity of nanostructre
    
    For a given incident field, calculate the electro-magnetic field in the 
    proximity of the nanostructure (positions defined by `MAP`).
    
    - Pure python implementation.
    
    - CUDA support to run on GPU, which can be significantly faster on large problems.
    
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    r_probe : tuple (x,y,z) or list of 3-lists/-tuples
        (list of) coordinate(s) to evaluate nearfield on. 
        Format: tuple (x,y,z) or list of 3 lists: [Xmap, Ymap, Zmap] 
        (the latter can be generated e.g. using :func:`.tools.generate_NF_map`)
        
    which_fields : list of str, default: ["Es","Et","Bs","Bt"]
        which fields to calculate and return. available options: 
            ["Es","Et","E0", "Bs","Bt","B0"]

    val_inside_struct : str, default: "field"
        one of ["field", "0", "zero", "NaN", "none", None].
        value to return for positions inside structure. default "field" returns 
        the field at the location of the closest meshpoint. Can be very coarse.
        Disable by setting "None", but note that inside the structure the 
        Green's function will diverge in the latter case.

    val_outside_struct : str, default: "field"
        see `val_inside_struct`
    
    N_neighbors_internal_field : int, default: 4
        Average internal field over field at *N* closes meshpoints. Neighbor 
        fields are weighted by the distance of the evaluation point to 
        the respective neighbor mesh cell.
    
    method : str
        deprecated, has no effect
        
    
    Returns
    -------
    depending on kwarg `which_fields`, up to 4 lists of 6-tuples, complex : 
        - scattered Efield ("Es")
        - total Efield ("Et", inlcuding fundamental field)
        - scattered Bfield ("Bs")
        - total Bfield ("Bt", inlcuding fundamental field)
    
    the tuples are of shape (X,Y,Z, Ax,Ay,Az) with Ai the corresponding 
    complex field component
        
    
    Notes
    -----
    For details of the calculation of the scattered field outside the 
    nano-object using the self-consistent field inside the particle, see e.g.: 
    Girard, C. **Near fields in nanostructures.** 
    Reports on Progress in Physics 68, 1883–1933 (2005).
        
    """
# =============================================================================
#     Exception handling
# =============================================================================
    if str(val_inside_struct).lower() == 'none':
        try:
            import scipy
            if int(scipy.__version__.split('.')[0]) == 0 and int(scipy.__version__.split('.')[1]) < 17:
                raise Exception("scipy with version < 0.17.0 installed! " +
                                "Positions inside nanostructure cannot be " +
                                "identified. Please upgrade or set `val_inside_struct`=None.")
        except ImportError:
            raise Exception("It seems scipy is not installed. Scipy is required " +
                            "by `nearfield` for detecting internal field positions. " +
                            "Please install scipy >=v0.17, or set `val_inside_struct`=None.")
        
    which_fields = [wf.lower() for wf in which_fields]
    
# =============================================================================
#     preparation
# =============================================================================
    from pyGDM2 import tools


    if len(np.shape(r_probe)) == 1:
        if len(r_probe) == 3:
            r_probe = [[r_probe[0]], [r_probe[1]], [r_probe[2]]]
        else: 
            raise ValueError("If 'r_probe' is tuple, must consist of *exactly* 3 elements!")
    elif len(np.shape(r_probe)) == 2:
        if np.shape(r_probe)[0] != 3 and np.shape(r_probe)[1] != 3:
            raise ValueError("'r_probe' must consist of *exactly* 3 elements!")
        if np.shape(r_probe)[0] != 3:
            r_probe = np.transpose(r_probe)
    else:
        raise ValueError("wrong format for 'r_probe'. must consist of *exactly* 3 " +
                         "elements, either floats, or lists.")
    r_probe = np.transpose(r_probe)
    
    field_kwargs    = tools.get_field_indices(sim)[field_index]
    wavelength      = field_kwargs['wavelength']
    
    sim.struct.setWavelength(wavelength)
    conf_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    
    
    ## --- fundamental field evaluation - use dummy structure with 
    env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    E0 = sim.efield.field_generator(r_probe, env_dict, **field_kwargs)
    B0 = sim.efield.field_generator(r_probe, env_dict, returnField='H', **field_kwargs)
    
    
    ## --- electric polarization of each discretization cell via tensorial polarizability
    if "es" in which_fields or "et" in which_fields or \
                "bs" in which_fields or "bt" in which_fields or \
                "hs" in which_fields or "ht" in which_fields:
        if sim.E is None: 
            raise ValueError("Error: Scattered field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
        Eint = sim.E[field_index][1]
        alpha_tensor = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)
        P = np.matmul(alpha_tensor, Eint[...,None])[...,0]
    
        
        ## --- init empty numpy arrays
        if "es" in which_fields or "et" in which_fields:
            Escat = np.zeros(shape=(len(r_probe), 3), dtype=sim.efield.dtypec)
        if "bs" in which_fields or "bt" in which_fields or "hs" in which_fields or "ht" in which_fields:
            Bscat = np.zeros(shape=(len(r_probe), 3), dtype=sim.efield.dtypec)
        
# =============================================================================
#     treat positions inside structure
# =============================================================================
        val_inside_struct = str(val_inside_struct).lower()
        
        _i_r_probe_out = []
        _i_r_probe_in = []
        _r_probe_out = []
        if val_inside_struct.lower() != "none":
            from scipy.linalg import norm
            for i, R in enumerate(r_probe):
                dist_list = norm(sim.struct.geometry - R, axis=1)
                idcs_min_dist = np.argsort(dist_list)
                ## --- if inside, replace fields
                if abs(dist_list[idcs_min_dist[0]]) <= 1.005*sim.struct.step:
                    if val_inside_struct.lower() == "nan":
                        fill_valueE = np.nan
                        fill_valueB = np.nan
                    elif val_inside_struct.lower() == "field":  # calc internal scattered field
                        fill_valueE = np.average(Eint[idcs_min_dist[:N_neighbors_internal_field]], 
                                     weights = 1/(sim.struct.step/100. + dist_list[idcs_min_dist[:N_neighbors_internal_field]]**1), 
                                     axis=0) - E0[i]  # subtract E0 --> scattered field only
                        if sim.H is not None:
                            Hint = sim.H[field_index][1]
                            fill_valueB = np.average(Hint[idcs_min_dist[:N_neighbors_internal_field]], 
                                     weights = 1/(sim.struct.step/100. + dist_list[idcs_min_dist[:N_neighbors_internal_field]]**1), 
                                     axis=0) - B0[i]  # subtract B0 --> scattered field only
                        else:
                            if "bs" in which_fields or "bt" in which_fields or "hs" in which_fields or "ht" in which_fields:
                                warnings.warn("No H-values inside structure calculated. Setting internal magnetic field zero. " + 
                                              "Please run 'scatter' with the according parameter.")
                            fill_valueB = 0   # here we need the B-field inside. Not yet implemented
                    else:
                        fill_valueE = 0
                        fill_valueB = 0
                    if "es" in which_fields or "et" in which_fields:
                        Escat[i] = fill_valueE
                    if "bs" in which_fields or "bt" in which_fields or "hs" in which_fields or "ht" in which_fields:
                        Bscat[i] = fill_valueB
                        
                    _i_r_probe_in.append(i)
                ## --- not inside, requires field propagation
                else:
                    _i_r_probe_out.append(i)
                    _r_probe_out.append(R)
        
        _r_probe_out = np.array(_r_probe_out)
        _i_r_probe_out = np.array(_i_r_probe_out)
        _i_r_probe_in = np.array(_i_r_probe_in)
            
# =============================================================================
#     evaluate Green's function and propagate fields
# =============================================================================
        val_outside_struct = str(val_outside_struct).lower()
        
        if len(_r_probe_out) > 0:
            ## +++++++++++++ electric field +++++++++++++
            if "es" in which_fields or "et" in which_fields:
                if val_outside_struct == 'field':
                    ## Greens function for each dipole
                    G_NF = np.zeros((len(sim.struct.geometry), len(_r_probe_out), 3, 3), 
                                    dtype = sim.efield.dtypec)
                    sim.dyads.eval_G(sim.struct.geometry, _r_probe_out, 
                                     sim.dyads.G_EE, wavelength, conf_dict, G_NF)
                    
                    ## propagate fields 
                    G_NF = np.moveaxis(G_NF, 0,2).reshape((len(_r_probe_out), 3, -1))  # re-arrange for direct matrix-vector multiplication
                    _Escat_out = np.zeros(shape=(len(_r_probe_out), 3), dtype=sim.efield.dtypec)
                    _calc_repropagation(P, _Escat_out, G_NF)
                    
                    ##  fill final field-list with outside values
                    Escat[_i_r_probe_out] = _Escat_out
                elif val_outside_struct == "nan":
                    Escat[_i_r_probe_out] = np.nan
                else:
                    Escat[_i_r_probe_out] = 0
                    
            ## +++++++++++++ magnetic field +++++++++++++
            if ("bs" in which_fields or "bt" in which_fields or "hs" in which_fields 
                    or "ht" in which_fields):
                if val_outside_struct == 'field':
                    ## Greens function for each dipole
                    G_NF = np.zeros((len(sim.struct.geometry), len(_r_probe_out), 3, 3), 
                                    dtype = sim.efield.dtypec)
                    sim.dyads.eval_G(sim.struct.geometry, _r_probe_out, 
                                     sim.dyads.G_HE, wavelength, conf_dict, G_NF)
                    
                    ## propagate fields 
                    G_NF = np.moveaxis(G_NF, 0,2).reshape((len(_r_probe_out), 3, -1))  # re-arrange for direct matrix-vector multiplication
                    _Bscat_out = np.zeros(shape=(len(_r_probe_out), 3), dtype=sim.efield.dtypec)
                    _calc_repropagation(P, _Bscat_out, G_NF)
                    
                    ##  fill final field-list with outside values
                    Bscat[_i_r_probe_out] = _Bscat_out
                elif val_outside_struct == "nan":
                    Bscat[_i_r_probe_out] = np.nan
                else:
                    Bscat[_i_r_probe_out] = 0
    
# =============================================================================
#     bundle output
# =============================================================================
        if "et" in which_fields:
            Etot = Escat.copy()
            if val_outside_struct == 'field' and len(_i_r_probe_out)>0:
                Etot[_i_r_probe_out] = Etot[_i_r_probe_out] + E0[_i_r_probe_out]
            if val_inside_struct == 'field' and len(_i_r_probe_in)>0:
                Etot[_i_r_probe_in] = Etot[_i_r_probe_in] + E0[_i_r_probe_in]
        if "bt" in which_fields or "ht" in which_fields:
            Btot = Bscat.copy()
            if val_outside_struct == 'field' and len(_i_r_probe_out)>0:
                Btot[_i_r_probe_out] = Btot[_i_r_probe_out] + B0[_i_r_probe_out]
            if val_inside_struct == 'field' and len(_i_r_probe_in)>0:
                Btot[_i_r_probe_in] = Btot[_i_r_probe_in] + B0[_i_r_probe_in]
    
    
    return_field_list = []
    for f_type in which_fields:
        
        if f_type.lower() == "es":
            Escat = np.concatenate([r_probe, Escat], axis=1)
            return_field_list.append(Escat)
        if f_type.lower() == "et":
            Etot = np.concatenate([r_probe, Etot], axis=1)
            return_field_list.append(Etot)
        if f_type.lower() == "e0":
            E0 = np.concatenate([r_probe, E0], axis=1)
            return_field_list.append(E0)
            
        if f_type.lower() in ["bs", "hs"]:
            Bscat = np.concatenate([r_probe, Bscat], axis=1)
            return_field_list.append(Bscat)
        if f_type.lower() in ["bt", "ht"]:
            Btot = np.concatenate([r_probe, Btot], axis=1)
            return_field_list.append(Btot)
        if f_type.lower() in ["b0", "h0"]:
            B0 = np.concatenate([r_probe, B0], axis=1)
            return_field_list.append(B0)
    
    return return_field_list



def internal_field_intensity(sim, field_index, which_value='average',
                             which_field='E'):
    """return internal field intensity
    
    returns either average or peak (min or max) internal field intensity for 
    either electric of magnetic field.
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
        
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    which_value : str, default: 'average'
        - if 'average', 'avg' or 'mean': average field intensity
        - if 'max' or 'maximum': maximum field intensity
        - if 'min' or 'minimum': minimum field intensity
    
    which_field : str, default: 'E'
        one of 'E', 'H' or 'both' (either electric or magnetic field or both)
    
    """
    if sim.E is None: 
        raise ValueError("Error: Scattered field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
    
    ## --- E field
    if which_field.lower() in ['e', 'both']:
        if sim.E is not None:
            E = sim.E[field_index][1]
            I_E_mean = np.sum(np.abs(E)**2, axis=1).mean()
        else:
            raise Exception("E field not evaluated. Please run 'sim.scatter()'.")
    
    ## --- H field
    if which_field.lower() in ['h', 'b', 'both']:
        if sim.H is not None:
            H = sim.H[field_index][1]
            I_H_mean = np.sum(np.abs(H)**2, axis=1).mean()
        else:
            raise Exception("H field not evaluated. Please run 'sim.scatter(calc_H=True)'.")
    
    if which_field.lower() in ['h', 'b']:
        return I_H_mean
    elif which_field.lower() in ['e']:
        return I_E_mean
    else:
        return I_E_mean, I_H_mean



def poynting(sim, field_index, r_probe, which_fields="tot"):
    """calculate the time average Poynting vector
    
    calculate the time average Poynting vector of the scattered, total or 
    fundamental (incident) field of a pyGDM simulation at one or more locations
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    r_probe : tuple (x,y,z) or list of 3-lists/-tuples
        (list of) coordinate(s) to evaluate nearfield on. 
        Format: tuple (x,y,z) or list of 3 lists: [Xmap, Ymap, Zmap] 
        (the latter can be generated e.g. using :func:`.tools.generate_NF_map`)
    
    which_fields : str, default: "tot"
        which type of fields to use. one of: "scat", "tot", "0".
        scattered field, total field (=E_scat + E_0) or incident field alone.
        
        
    Returns
    -------
    P : list of 6-tuples (x,y,z, Px, Py, Pz)
    """
    
    if "scat" == which_fields.lower():
        wf= ['es', 'hs']
    if "tot" == which_fields.lower():
        wf= ['et', 'ht']
    if "0" == which_fields.lower() or "in" == which_fields.lower():
        wf= ['e0', 'h0']
    
    E, H = nearfield(sim, field_index, r_probe, which_fields=wf)
    
    P = np.copy(E)
    P[:,3:] = np.cross(np.conj(E[:,3:]), H[:,3:])

    return P.real


# =============================================================================
# field gradients
# =============================================================================
def field_gradient(sim, field_index, 
                   r_probe=None, delta=None, which_fields=["Et"]):
    """nearfield gradient distribution inside or in proximity of nanostructure
    
    Calculate field-gradients (positions defined by `r_probe`). 
    Using center difference for the numerical derivatives.
    
    Implemented by C. Majorel.
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    r_probe : tuple (x,y,z) or list of 3-lists/-tuples, default: positions inside geometry
        (list of) coordinate(s) to evaluate nearfield on. 
        Format: tuple (x,y,z) or list of 3 lists: [Xmap, Ymap, Zmap] 
        (the latter can be generated e.g. using :func:`.tools.generate_NF_map`)
        If `None`: calculate inside nanostructure
    
    delta : float, default=None
        differential step for numerical center-derivative (in nanometers).
        If `None`, use stepsize, which is recommended inside a nanostructure.
        Finer steps are recommended for fields outside the nanostructure
        (1nm may be a good choice).
        
    which_fields : list of str, default: ["Et"]
        for which fields to calculate the gradient. available options: 
            ["Es","Et","E0", "Hs","Ht,"H0]  ("B" can be used as equivalent to "H")
    
    Returns
    -------
    depending on kwarg `which_gradfields`, up to 6 lists of 3 lists of 6-tuples, complex : 
        - gradient scattered Efield ("Es")
        - gradient total Efield ("Et", includes incident field)
        - gradient incident Efield ("E0")
        - gradient scattered Hfield ("Hs")
        - gradient total Hfield ("Ht", includes incident field)
        - gradient incident Hfield ("H0")
    
    each of the lists contains 3 lists [dAx, dAy, dAz] with dAj the differential terms
    dE[0] = dE/dx = [dE0x/dx, dE0y/dx, dE0z/dx]
    dE[1] = dE/dy = [dE0x/dy, dE0y/dy, dE0z/dy]
    dE[2] = dE/dz = [dE0x/dz, dE0y/dz, dE0z/dz]
        
    """
# =============================================================================
#     Exception handling
# =============================================================================
    if r_probe is None:
        r_probe = sim.struct.geometry
    if delta is None:
        delta= sim.struct.step
    
    
    if len(r_probe) == 0:
        return np.zeros((3,0,6))
    if len(np.shape(r_probe)) == 1:
        if len(r_probe) == 3:
            r_probe = [[r_probe[0]], [r_probe[1]], [r_probe[2]]]
        else: 
            raise ValueError("If 'r_probe' is tuple, must consist of *exactly* 3 elements!")
    elif len(np.shape(r_probe)) == 2:
        if np.shape(r_probe)[0] != 3 and np.shape(r_probe)[1] != 3:
            raise ValueError("'r_probe' must consist of *exactly* 3 elements!")
        if np.shape(r_probe)[0] != 3:
            r_probe = np.transpose(r_probe)
    else:
        raise ValueError("wrong format for 'r_probe'. must consist of *exactly* 3 " +
                         "elements, either floats, or lists.")
    r_probe = np.transpose(r_probe)
    
    
# =============================================================================
#     preparation
# =============================================================================
    from pyGDM2 import tools
    
    which_fields = [wf.lower() for wf in which_fields]
    
    field_kwargs = tools.get_field_indices(sim)[field_index]
    wavelength = field_kwargs['wavelength']
    sim.struct.setWavelength(wavelength)    
    
    
# =============================================================================
#     calc field gradients
# =============================================================================
    ## --- d/dx    
    r_probe_px = copy.deepcopy(r_probe)     # Xmap+delta
    r_probe_mx = copy.deepcopy(r_probe)     # Xmap-delta
    r_probe_px.T[0] += delta
    r_probe_mx.T[0] -= delta
    F_px = nearfield(sim, field_index, r_probe_px, which_fields)
    F_mx = nearfield(sim, field_index, r_probe_mx, which_fields)

    dFdx = []
    for Epx, Emx in zip(F_px, F_mx):
        dFdx.append(np.array([np.ravel((Epx.T[3:].T - Emx.T[3:].T) / (2.*delta))]))

    ## --- d/dy
    r_probe_py = copy.deepcopy(r_probe)     # Ymap+delta
    r_probe_my = copy.deepcopy(r_probe)     # Ymap-delta
    r_probe_py.T[1] += delta
    r_probe_my.T[1] -= delta
    F_py = nearfield(sim, field_index, r_probe_py, which_fields)
    F_my = nearfield(sim, field_index, r_probe_my, which_fields)
    
    dFdy = []
    for Epy, Emy in zip(F_py, F_my):
        dFdy.append(np.array([np.ravel((Epy.T[3:].T - Emy.T[3:].T) / (2.*delta))]))
    
    ## --- d/dz    
    r_probe_pz = copy.deepcopy(r_probe)     # Zmap+delta
    r_probe_mz = copy.deepcopy(r_probe)     # Zmap-delta
    r_probe_pz.T[2] += delta
    r_probe_mz.T[2] -= delta
    F_pz = nearfield(sim, field_index, r_probe_pz, which_fields)
    F_mz = nearfield(sim, field_index, r_probe_mz, which_fields)
    
    ## --- assemble
    dFdz = []
    for Epz, Emz in zip(F_pz, F_mz):
        dFdz.append(np.array([np.ravel((Epz.T[3:].T - Emz.T[3:].T) / (2.*delta))]))
        
        
    return_field_list = []
    for Fdx, Fdy, Fdz in zip(dFdx, dFdy, dFdz):
        ddx = np.concatenate([r_probe, np.reshape(Fdx, (len(r_probe), 3))], axis=1)
        ddy = np.concatenate([r_probe, np.reshape(Fdy, (len(r_probe), 3))], axis=1)
        ddz = np.concatenate([r_probe, np.reshape(Fdz, (len(r_probe), 3))], axis=1)
        return_field_list.append([ddx.copy(), ddy.copy(), ddz.copy()])
    
    if len(return_field_list) == 1:
        return_field_list = return_field_list[0]
    return return_field_list





def optical_force(sim, field_index, 
                  return_value='total'):
    """optical force acting on the nanostructure
    
    ** ------- FUNCTION STILL UNDER TESTING ------- **
    
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    return_value : str, default: 'total'
        Values to be returned. Either 'total' (default) or 'structure'.
        
        - "total" : return the total force on the entire particle (3-vector [Fx, Fy, Fz])
        
        - "structure" : return spatially resolved force, in other words 
        the force acting on each meshpoint. --> list of tuples [x,y,z,F_ix,F_iy,F_iz])
    
    
    Returns
    -------
    F : tuple (Fx,Fy,Fz) *or* list of force vectors [x,y,z,Fx,Fy,Fz]
    
        - `return_value`=="total" : (3-vector [Fx, Fy, Fz])
          total force on structure 

        - `return_value`=="structure" : (return list of tuples [xi,yi,zi,F_ix,F_iy,F_iz])
          force at all mesh-cells `i` of positions [xi,yi,zi].
    
    
    Notes
    -----
    For details on optical force calculations see e.g.:
    
    [1] Chaumet & Nieto-Vesperinas. 
    *Coupled dipole method determination of the electromagnetic force 
    on a particle over a flat dielectric substrate*. 
    Phys. Rev. B 61, 14119–14127 (2000)
        
    """
    if sim.E is None: 
        raise ValueError("Error: Scattered field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
    
    warnings.warn("Optical force calculation is a beta-functionality and still under testing. " +
                  "Please use with caution.")
    
    ## -- electric polarization of structure 
    Eint = sim.E[field_index][1]
    alpha_tensor = sim.dyads.getPolarizabilityTensor(sim.E[field_index][0]['wavelength'], 
                                                     sim.struct)
    P = np.matmul(alpha_tensor, Eint[...,None])[...,0]
    
    
    ## -- field gradients inside half structure 
    gradE_inside = field_gradient(sim, field_index)
    
    ## -- separate the derivatives
    dEdx = gradE_inside[0][...,3:]       #  [X, Y, Z, dEX/dx, dEY/dx, dEZ/dx] --> [dEX/dx, dEY/dx, dEZ/dx]
    dEdy = gradE_inside[1][...,3:]       # --> [dEX/dy, dEY/dy, dEZ/dy]
    dEdz = gradE_inside[2][...,3:]       # --> [dEX/dz, dEY/dz, dEZ/dz]
    
    ## -- calculate the force
    Fx = 0.5 * np.real( P[...,0] * np.conj(dEdx[...,0]) + 
                        P[...,1] * np.conj(dEdx[...,1]) + 
                        P[...,2] * np.conj(dEdx[...,2]) )
    Fy = 0.5 * np.real( P[...,0] * np.conj(dEdy[...,0]) + 
                        P[...,1] * np.conj(dEdy[...,1]) + 
                        P[...,2] * np.conj(dEdy[...,2]) )
    Fz = 0.5 * np.real( P[...,0] * np.conj(dEdz[...,0]) + 
                        P[...,1] * np.conj(dEdz[...,1]) + 
                        P[...,2] * np.conj(dEdz[...,2]) )
    
    ## -- return results
    if return_value.lower() == "structure":
        np.transpose([sim.struct.geometry.T[0], 
                      sim.struct.geometry.T[1], 
                      sim.struct.geometry.T[2], 
                      np.sum(Fx), np.sum(Fy), np.sum(Fz)])
    else:
        F_tot = np.array([np.sum(Fx), np.sum(Fy), np.sum(Fz)])
        return F_tot




def optical_chirality(sim, field_index, r_probe, which_field="t", **kwargs):
    """calculate the optical chirality of the electromagnetic field
    
    ** ------- FUNCTION STILL UNDER TESTING ------- **
    
    Returns the normalized electromagnetic chirality *C / C_LCP*, as defined in [1-4].
    Normalized to a left circular polarized (LCP) plane wave of amplitude |E_0|=1.
    Hence:
    
     --> LCP: C = + 1
     --> RCP: C = - 1
    
    see also [3] or [4] for a short discussion about the normalization to LCP.
    
    |C/C_LCP|>1 means that the local field is "superchiral", 
    hence a chiral molecule is excited with higher selectivity than it would be
    with circular polarized light.
    
    kwargs are passed to :func:`.nearfield`.
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    r_probe : tuple (x,y,z) or list of 3-lists/-tuples
        (list of) coordinate(s) to evaluate nearfield on. 
        Format: tuple (x,y,z) or list of 3 lists: [Xmap, Ymap, Zmap] 
        (the latter can be generated e.g. using :func:`.tools.generate_NF_map`)
    
    which_field : str, default: 't'
        either of:
            - "s", "Es" for *scattered field*
            - "t", "Et" for *total field* (with illumination, i.e. Et=Es+E0; Bt=Bs+B0)
            - "0", "E0" for *incident field* (illumination (E0, B0) only, corresponding to simulation with no nanostructure)
    
    Returns
    -------
    list of tuples (X,Y,Z, X). Each tuple consists of the position and the 
    (normalized) optical chirality *C / C_LCP*.
    
    
    Notes
    -----
    For details on the optical chirality, see e.g.:
    
    [1] Tang, Y. & Cohen, A. E.: **Optical Chirality and Its Interaction 
    with Matter**. Phys. Rev. Lett. 104, 163901 (2010)
    
    [2] Meinzer, N., Hendry, E. & Barnes, W. L.: **Probing the chiral nature 
    of electromagnetic fields surrounding plasmonic nanostructures**. 
    Phys. Rev. B 88, 041407 (2013)
    
    A discussion about the proper normalization of C can be found in:
    
    [3] 1.Schäferling, M., Yin, X.,  Giessen, H. 
    **Formation of chiral fields in a symmetric environment**. 
    Opt. Express 20(24), 26326–26336 (2012)

    [4] Schäferling M., Yin X., Engheta N., Giessen H. **Helical 
    Plasmonic Nanostructures as Prototypical Chiral Near-Field Sources**. 
    ACS Photonics 1(6), 530-537 (2014)
    
    """
    if sim.E is None: 
        raise ValueError("Error: Scattered field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
    
    Es, Et, E0, Bs, Bt, B0 = nearfield(sim, field_index, r_probe, 
                                       which_fields=['es', 'et', 'e0', 'bs', 'bt', 'b0'], 
                                       **kwargs)
    
    if which_field.lower() in ["s", "es", "scat", "scattered"]:
        Et = Es
        Bt = Bs
    elif which_field.lower() in ["0", "e0"]:
        Et = E0
        Bt = B0
    
    ## Ref. [2]: C ~ Im(E* B)
    C = np.concatenate([
                    Et.T[:3].real, # positions
                    [-1*np.sum(np.multiply(np.conjugate(Et.T[3:]), 
                                                     Bt.T[3:]), axis=0).imag]
                        ]).astype(np.float32)
    
    if len(np.shape(r_probe)) == 1 or len(r_probe) == 1:
        return C[3][0]
    else:
        return C
    


# =============================================================================
# heat and temperature
# =============================================================================
def heat(sim, field_index, power_scaling_e0=1.0, return_value='total', return_units='nw'):
    """calculate the total induced heat in the nanostructure
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    power_scaling_e0 : float, default: 1
        incident laser power scaling. power_scaling_e0 = 1 corresponds 
        to 1 mW per micron^2. See [1].
    
    return_value : str, default: 'total'
        Values to be returned. Either 'total' (default) or 'structure'.
        
        - "total" : return the total deposited heat in nW (float)
        
        - "structure" : return spatially resolved deposited heat at each 
          meshpoint in nW (list of tuples [x,y,z,q])
    
    return_units : str, default: "nw"
        units of returned heat values, either "nw" or "uw" (nano or micro Watts)
    
    
    Returns
    -------
    Q : float *or* list of tuples [x,y,z,q]
    
        - `return_value`=="total" : (return float)
          total deposited heat in nanowatt (optionally in microwatt). 

        - `return_value`=="structure" : (return list of tuples [x,y,z,q])
          The returned quantity *q* is the total deposited heat 
          at mesh-cell position [x,y,z] in nanowatt. To get the heating 
          power-density, please divide by the mesh-cell volume.
    
    
    Notes
    -----
    For details on heat/temperature calculations and raster-scan simulations, see:
    
    [1] Baffou, G., Quidant, R. & Girard, C.: **Heat generation in plasmonic 
    nanostructures: Influence of morphology**
    Applied Physics Letters 94, 153109 (2009)
    
    [2] Teulle, A. et al.: **Scanning optical microscopy modeling in nanoplasmonics** 
    Journal of the Optical Society of America B 29, 2431 (2012).


    """
    if sim.E is None: 
        raise ValueError("Error: Scattering field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
    
    
    power_scaling_e0 *= 0.01    # for mW/cm^2
    
    field_params = sim.E[field_index][0]
    wavelength   = field_params['wavelength']
    
    ## --- Factor allowing to have the released power in nanowatt 
    released_power_scaling = 100.
    
    ## --- polarizabilities and electric fields
    sim.struct.setWavelength(wavelength)
    n_env        = sim.dyads.getEnvironmentIndices(wavelength, sim.struct.geometry[:1])[0]  # assume structure is fully in one environment
    alpha_tensor = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)
    E = sim.E[field_index][1]
    P = np.matmul(alpha_tensor, E[...,None])[...,0]

    
    ## heat at each meshpoint in nW/nm^3
    q = (((8 * np.pi**2 / wavelength) / n_env) * np.sum(np.multiply(P, np.conjugate(E)), axis=1).imag
                         * power_scaling_e0 * released_power_scaling).real
    
    ## --- optional conversion to micro watts
    if return_units.lower() == 'uw':
        q /= 1.0E3
    
    if return_value == 'total':
        return np.sum(q)
    elif return_value in ['structure', 'struct']:
        x,y,z = sim.struct.geometry.T
        return np.concatenate([[x],[y],[z],[q]]).T
    else:
        raise ValueError("`return_value` must be one of ['total', 'structure', 'struct'].")



def temperature(sim, field_index, r_probe, 
                kappa_env=0.6, kappa_subst=None, incident_power=1.0):
    """calculate the temperature rise at locations outside the nano-particle
    
    Calculate the temperature increase close to a optically excited 
    nanostructure using the approach described in [2] and [3] (Ref. [3] 
    introduces a further correction term for particles lying on a substrate)
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    r_probe : tuple (x,y,z) or list of 3-lists/-tuples
        (list of) coordinate(s) to evaluate nearfield on. 
        Format: tuple (x,y,z) or list of 3 lists: [Xmap, Ymap, Zmap] 
        (the latter can be generated e.g. using :func:`.tools.generate_NF_map`)
    
    kappa_env : float, default: 0.6
        heat conductivity of environment. default: kappa_env = 0.6 (water). 
        (air: kappa=0.024, for more material values see e.g. [4]). In W/mK.
    
    kappa_subst : float, default: None
        heat conductivity of substrate. default: None --> same as substrate. 
        Using the mirror-image technique described in [3]. (glass: kappa=0.8)
    
    incident_power : float, default: 1.0
        incident laser power density in mW per micron^2. See also [1].
    
    
    Returns
    -------
    if evaluating at a single position, D_T : float
        temperature increase in Kelvin at r_probe
    
    if evaluating at a list of positions, list of tuples [x, y, z, D_T] 
        where D_T is the temperature increase in Kelvin at (x,y,z), which
        are the positions defined by `r_probe`
    
    Notes
    -----
    For details on heat/temperature calculations and raster-scan simulations, see:
    
    [1] Baffou, G., Quidant, R. & Girard, C.: **Heat generation in plasmonic 
    nanostructures: Influence of morphology**
    Applied Physics Letters 94, 153109 (2009)
    
    [2] Baffou, G., Quidant, R. & Girard, C.: **Thermoplasmonics modeling: 
    A Green’s function approach** 
    Phys. Rev. B 82, 165424 (2010)

    [3] Teulle, A. et al.: **Scanning optical microscopy modeling in nanoplasmonics**
    Journal of the Optical Society of America B 29, 2431 (2012).
    
    
    For the thermal conductivity of common materials, see:
    
    [4] Hugh D Young, Francis Weston Sears: **University Physics**, *chapter 15*,
    8th. edition: Addison-Wesley, 1992
    (see also e.g.: http://hyperphysics.phy-astr.gsu.edu/hbase/Tables/thrcn.html)

    """
    kappa_subst  = kappa_subst or kappa_env
    field_params = sim.E[field_index][0]
    wavelength   = field_params['wavelength']
    sim.struct.setWavelength(wavelength)
    n_env        = sim.dyads.getEnvironmentIndices(wavelength, sim.struct.geometry[:1])[0]  # assume structure is fully in one environment
    
    if sim.E is None: 
        raise ValueError("Error: Scattering field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
    
    if kappa_subst == kappa_env and sim.dyads.n1_material.__name__ != sim.dyads.n2_material.__name__:
        warnings.warn("Substrate and environment have different indices but same heat conductivity.")
    if kappa_subst != kappa_env and sim.dyads.n1_material.__name__ == sim.dyads.n2_material.__name__:
        warnings.warn("Substrate and environment have same ref.index but different heat conductivity.")
    
    incident_power *= 0.01          # for mW/cm^2
    released_power_scaling = 100.   # output in nW
    
    ## --- main heat generation function
    def calc_heat_single_position(sim, r_probe):
        ## --- polarizability and field in structure
        alpha_tensor = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)
        E = sim.E[field_index][1]
        P = np.matmul(alpha_tensor, E[...,None])[...,0]
        
        ## --- meshpoint distance to probe, polarizabilities and electric fields
        r_mesh = sim.struct.geometry
        dist_probe = np.sqrt( np.sum( np.power((np.array(r_probe) - r_mesh), 2), axis=1) )
        
        ## --- mirror structure below substrate for heat reflection at substrate
        if kappa_subst != kappa_env:
            r_mesh_mirror = copy.deepcopy(sim.struct.geometry)
            r_mesh_mirror.T[2] *= -1
            dist_probe_mirror = np.sqrt( np.sum( np.power((np.array(r_probe) - r_mesh_mirror), 2), axis=1) ) 
        
        ## --- temperature rise at r_probe
        q = (((2 * np.pi / wavelength) / n_env) * np.sum(np.multiply(P, np.conjugate(E)), axis=1).imag
                         * incident_power * released_power_scaling).real
        D_T = np.sum( q / dist_probe )
        if kappa_subst != kappa_env:
            D_T += np.sum( q / dist_probe_mirror ) * (kappa_subst - kappa_env) / (kappa_subst + kappa_env)
        D_T /= kappa_env
        
        return D_T
    
    
    ## --- single position:
    if len(np.shape(r_probe)) == 1 or len(r_probe) == 1:
        D_T = calc_heat_single_position(sim, r_probe)
    ## --- multiple positions:
    else:
        D_T = []
        if np.shape(r_probe)[1] != 3:
            r_probe = np.transpose(r_probe)
        for i in r_probe:
            D_T.append([i[0], i[1], i[2], 
                        calc_heat_single_position(sim, i)])
        D_T = np.array(D_T)
    
    return D_T
























# =============================================================================
# deprecations - to be removed in a future version
# =============================================================================
def multipole_decomp(sim, field_index, r0=None, quadrupoles=False):
    """multipole decomposition of nanostructure optical response
    
    ** ------- DEPRECATED ------- **
    
    please use :func:`pyGDM2.multipole.multipole_decomposition_exact`
    
    """
    from pyGDM2 import multipole
    
    warnings.warn("This function is deprecated and will be removed in a future version. " +
                  "Please use the `pyGDM2.multipole` module instead.")
    
    if quadrupoles:
        p, m, q, mq = multipole.multipole_decomposition_exact(
                                    sim, field_index, r0=r0, epsilon=0.01, 
                                    which_moments=['p', 'm', 'qe', 'qm'])
        return p, m, q, mq

    else:
        p, m = multipole.multipole_decomposition_exact(
                                    sim, field_index, r0=r0, epsilon=0.01, 
                                    which_moments=['p', 'm'])
        return p, m


def multipole_decomp_extinct(sim, field_index, r0=None, eps_dd=0.1, quadrupoles=False):
    """extinction from multipole decomposition
    
    ** ------- DEPRECATED ------- **
    
    please use :func:`pyGDM2.multipole.multipole_decomposition_exact`
    """
    from pyGDM2 import multipole
    
    warnings.warn("This function is deprecated and will be removed in a future version. " +
                  "Please use the `pyGDM2.multipole` module instead.")
    
    sigma_ext_p, sigma_ext_m, sigma_ext_q, sigma_ext_mq = multipole.extinct(
                                                sim, field_index, r0=r0, eps_dd=eps_dd)
    if quadrupoles:
        return sigma_ext_p, sigma_ext_m, sigma_ext_q, sigma_ext_mq
    else:
        return sigma_ext_p, sigma_ext_m


def mutlipole_decomp(sim, field_index, r0=None):
    return multipole_decomp(sim, field_index, r0=r0, quadrupoles=True)

def mutlipole_decomp_extinct(sim, field_index, r0=None, eps_dd=0.1):
    return multipole_decomp_extinct(sim, field_index, r0, eps_dd, quadrupoles=True)



# encoding: utf-8
"""
Example script, demonstrating pyGDM multipole decomposition

"""
from __future__ import print_function, division

import numpy as np
import matplotlib.pyplot as plt

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields
from pyGDM2 import core
from pyGDM2 import visu
from pyGDM2 import tools
from pyGDM2 import linear


# =============================================================================
#
#                               !! warning !!
#
# multipole decomposition is not finally tested beta-functionality of pyGDM! 
# Use with caution and at own risk
# =============================================================================

## CUDA-based inversion on GPU
method = 'cupy'




## --- simulation initialization ---
mesh = 'cube'
step = 10
#geometry = structures.nanodisc(step, R=(175/step), H=int(280/step), mesh=mesh)
geometry = structures.rect_wire(step, L=int(np.round(160/step)), 
                                      W=int(np.round(160/step)), 
                                      H=int(np.round(160/step)), mesh=mesh)
material = materials.silicon()
norm = structures.get_normalization(mesh=mesh)
n1 = n2 = n3 = 1.0

struct = structures.struct(step, geometry, material, n1,n2, norm, n3=n3)

## incident field: lin. pol plane wave
field_generator = fields.planewave
wavelengths = np.linspace(530, 830, 31)
kwargs = dict(theta=0.0, kSign=-1)
efield = fields.efield(field_generator, 
               wavelengths=wavelengths, kwargs=kwargs)

## create simulation object
sim = core.simulation(struct, efield)


visu.structure(sim)
print("N dp={}".format(len(geometry)))

## --- run the main simulation ---
core.scatter(sim, matrix_setup='numba', method=method)


#%%
## -- spectra of extinction sections per multipole moment
wl, spec1 = tools.calculate_spectrum(sim, 0, linear.extinct)
ex, sc, ab = spec1.T
wl, spec2 = tools.calculate_spectrum(sim, 0, linear.mutlipole_decomp_extinct)
ex_p, ex_m, ex_q, ex_mq = spec2.T

#plt.figure(figsize=(10,7))
plt.plot(wl, ex, label='extinct')
plt.plot(wl, ex_p, label='p')
plt.plot(wl, ex_m, label='m')
plt.plot(wl, ex_q, label='q')
plt.plot(wl, ex_mq, label='mq')
plt.plot(wl, ex_p + ex_m + ex_q + ex_mq, label='multipole sum', dashes=[2,2])

plt.legend()
plt.title("multipole decomposition")
plt.xlabel("wavelength (nm)")
plt.ylabel("extinction cross section (nm^2)")
#plt.savefig("example_multipole_decomposition.png", dpi=150)
plt.show()

#%%
## -- spectra of multipole moments
wl, spec_dpdecomp = tools.calculate_spectrum(sim, 0, linear.mutlipole_decomp)
p = np.array([decomp[0] for decomp  in spec_dpdecomp])
m = np.array([decomp[1] for decomp  in spec_dpdecomp])
q = np.array([decomp[2] for decomp  in spec_dpdecomp])
mq = np.array([decomp[3] for decomp  in spec_dpdecomp])

plt.plot(wl, np.abs(p.T[0]), label='|p_x|')
plt.plot(wl, np.abs(m.T[1]), label='|m_y|')
plt.legend()
plt.xlabel("wavelength (nm)")
plt.ylabel("multipole moment (a.u.)")
plt.show()



# encoding: utf-8
"""
Created on May 6, 2017

@author: P. R. Wiecha

Example script demonstrating `pyGDM2.EO`:
    load best candidate of evolutionary optimizaton of plasmonic 
    nanostructure geometry for directional scattering

"""



import numpy as np
import matplotlib.pyplot as plt



from pyGDM2 import linear
from pyGDM2 import tools
from pyGDM2 import visu

from pyGDM2.EO.tools import get_best_candidate, get_best_candidate_f_x







## --- optimization file locations
results_folder = 'eo_out'
results_suffix = 'test_eo'




## --- reload and plot farfield intensity
N = 1
for i in range(1, N+1)[::-1]:    # reload and plot last N iterations
    
    ## --- load best candidate
    sim = get_best_candidate(results_folder, results_suffix, iteration=-1*i,
                             verbose=True)
    #tools.print_sim_info(sim)
    
    f, x =get_best_candidate_f_x(results_folder, results_suffix, iteration=-1*i)
    
    ## --- calculate farfield pattern
    tetalist, philist, I_sc, I_tot, I0 = linear.farfield(sim, field_index=0,
                                    r=10000, tetamin=0, tetamax=np.pi/2., 
                                    Nteta=18, Nphi=36)
#                                    Nteta=3, Nphi=5)
    
    
    ## --- plot
    plt.figure(figsize=(9,4))
    
    
    plt.subplot(121, aspect="equal")
    NF_imagX = tools.get_field_as_list_by_fieldindex(sim, 0).T[ [0,1,2,3] ]
    NF_imagX[3] = np.angle(NF_imagX[3].imag) * 180./np.pi
    
    plt.scatter(NF_imagX[0], NF_imagX[1], c = NF_imagX[3], cmap='bwr', marker='s')
    plt.colorbar(label=r'phase of $E_x$ (deg)')
    plt.clim([0, 180])
    
    
    
    plt.subplot(122, polar=True)
    visu.farfield_pattern_2D(tetalist, philist, I_sc, show=False)
    plt.colorbar()
    
    plt.show()














# encoding: utf-8
"""
Created on May 6, 2017

@author: P. R. Wiecha

Example script demonstrating `pyGDM2.EO`:
    Demonstrate how to implement a user-defined problem. 
    The example problem is the maximization of the scattering cross-section
    of a simple plasmonic nano-structure under plane wave illumination with 
    fixed wavelength and linear polarization angle.

"""

import numpy as np
import matplotlib.pyplot as plt

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields
from pyGDM2 import linear

from pyGDM2 import core

from pyGDM2.EO import models as eo_models
from pyGDM2.EO.core import do_eo


from PyGMO import algorithm



#==============================================================================
# Implement a simple problem: Maximize scattering cross-section
#==============================================================================
from pyGDM2.EO.problems import BaseProblem

class ownProblem(BaseProblem):
    """Problem to maximize scattering cross-section (SCS) of nano-structure"""
    
    def __init__(self, N_dim=9999, model=None, nthreads=1):
        """constructor"""
        super(self.__class__, self).__init__(N_dim, model, nthreads=nthreads)
    
    
    def objective_function(self, params):
        """evaluate directionality ratio
        """
        self.model.generate_structure(params)
        
        ## --- GDM simulation and cross-section calculation
        core.scatter(self.model.sim, verbose=0)
        ext_cs, sca_cs, abs_cs = linear.extinct(self.model.sim, 0)
        
        return float(sca_cs)      


    def human_readable_extra(self):
        return "\n\tMaximization of scattering cross section"







if __name__ == '__main__':
    #==============================================================================
    # Setup pyGDM part
    #==============================================================================
    ## ---------- Setup structure
    mesh = 'cube'
    step = 20
    material = materials.dummy(2.0)    # material: constant and real dielectric function
    material = materials.gold()        # material: gold
    n1, n2 = 1.0, 1.0         # constant environment
    
    ## --- Empty dummy-geometry, will be replaced on run-time by EO trial geometries
    geometry = []       
    
    struct = structures.struct(step, geometry, material, n1,n2, structures.get_normalization(mesh))
    
    
    
    ## ---------- Setup incident field
    field_generator = fields.planewave        # planwave excitation
    kwargs = dict(theta = [0.0])              # several polarizations
    wavelengths = [1200]                       # one single wavelength
    efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)
    
    
    
    ## ---------- Simulation initialization
    sim = core.simulation(struct, efield)
    
    
    
    
    #==============================================================================
    # setup evolutionary optimization
    #==============================================================================
    ## --- output folder and file-names
    results_folder = 'eo_out'
    results_suffix = 'test_own_problem'
    
    
    ## --- structure model and optimizaiton problem
    model = eo_models.BlockModel(sim, 
                                 20, 1,1,1, [-10,10],
                                 forbidden=[], symmetric=False)
    
    problem_dimension = model.get_dim()
    problem = ownProblem(problem_dimension, model)
    
    
    ## --- algorithm configuration
    algo = algorithm.jde(gen=1, memory=True)
    
    islands = 4             # number of islands ( > 1: multi-processing)
    population = 40         # individuals per island
    
    
    ## --- stop criteria
    max_time = 1800         # seconds
    max_iter = 200          # max. iterations
    max_nonsuccess = 20     # max. consecutive iterations without improvement
    
    
    
    
    
    #==============================================================================
    # run evolution
    #==============================================================================
    #do_eo(results_folder, results_suffix, problem)    # run with default params
    do_eo(results_folder, results_suffix,
          problem, algo, 
          islands, population,
          max_time, max_iter, max_nonsuccess, 
          ## --- flags
          runtime_plotting=True,
          save_each_generation=False,
          verbose=True)
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    



# encoding: utf-8
"""
Created on May 6, 2017

@author: P. R. Wiecha

Example script demonstrating `pyGDM2.EO`:
    Demonstrate how to implement a user-defined problem. 
    The example problem is the maximization of the scattering cross-section
    of a simple plasmonic nano-structure under plane wave illumination with 
    fixed wavelength and linear polarization angle.

"""

import numpy as np
import matplotlib.pyplot as plt

from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields

from pyGDM2 import core

from pyGDM2.EO.problems import ProblemMaxNearfield
from pyGDM2.EO.models import RectangularAntenna
from pyGDM2.EO.core import do_eo


from PyGMO import algorithm



#==============================================================================
# Setup pyGDM part
#==============================================================================
## ---------- Setup structure
mesh = 'cube'
step = 15
material = materials.gold()        # material: gold
n1, n2 = 1.0, 1.0         # constant environment

## --- Empty dummy-geometry, will be replaced on run-time by EO trial geometries
geometry = []       

struct = structures.struct(step, geometry, material, n1,n2, structures.get_normalization(mesh))



## ---------- Setup incident field
field_generator = fields.planewave        # planwave excitation
kwargs = dict(theta = [0.0])              # several polarizations
wavelengths = [800]                       # one single wavelength
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)



## ---------- Simulation initialization
sim = core.simulation(struct, efield)




#==============================================================================
# setup evolutionary optimization
#==============================================================================
## --- output folder and file-names
results_folder = 'eo_out'
results_suffix = 'test_max_NF_E'


## --- structure model and optimizaiton problem
limits_W   = [2, 20]
limits_L   = [2, 20] 
limits_pos = [-500, 500]
height = 1
model = RectangularAntenna(sim, limits_W, limits_L, limits_pos, height)

target = 'E'
r_probe = [0, 0, 80]
problem_dimension = model.get_dim()
problem = ProblemMaxNearfield(problem_dimension, model, r_probe=r_probe, opt_target=target)


## --- algorithm configuration
algo = algorithm.jde(gen=1, memory=True)

islands = 1             # number of islands ( > 1: multi-processing)
population = 20         # individuals per island


## --- stop criteria
max_time = 3600         # seconds
max_iter = 200          # max. iterations
max_nonsuccess = 30     # max. consecutive iterations without improvement





#==============================================================================
# run evolution
#==============================================================================
#do_eo(results_folder, results_suffix, problem)    # run with default params
do_eo(results_folder, results_suffix,
      problem, algo, 
      islands, population,
      max_time, max_iter, max_nonsuccess, 
      ## --- flags
      runtime_plotting=True,
      save_each_generation=False,
      verbose=True)













# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha, A. Arbouet, Y. Brûlé
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
"""
Collection of focused beam fields - hermite gauss, radial polarization, azimuthal polarization
"""

from __future__ import print_function
from __future__ import absolute_import

import multiprocessing
import warnings
import cmath
import numpy as np
import numba

import scipy.integrate as integrate
# from scipy import special
#import quadpy

## bessel functions 
# from scipy.special import jv, yv, j0
# from scipy.special import iv, kv



# =============================================================================
#               numba overloads for Bessel functions
#
# adapted for scipy>=1.4 from https://github.com/numba/numba-scipy
#
# =============================================================================
import ctypes
import scipy.special
from numba.extending import get_cython_function_address


## -- in future: add binding to complex hankel
## double complex hankel1(double, double complex)

name_to_numba_signatures = {
    'yv': [(numba.types.float64, numba.types.float64,), (numba.types.long_, numba.types.float64,)],
    'jv': [(numba.types.float64, numba.types.float64,), (numba.types.long_, numba.types.float64,)],
    
    'iv': [(numba.types.float64, numba.types.float64,), (numba.types.long_, numba.types.float64,)],
    'kv': [(numba.types.float64, numba.types.float64,), (numba.types.long_, numba.types.float64,)],
    }

name_and_types_to_pointer = {
    ('yv', numba.types.float64, numba.types.float64): ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double, ctypes.c_double)(get_cython_function_address('scipy.special.cython_special', '__pyx_fuse_1yv')),
    ('jv', numba.types.float64, numba.types.float64): ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double, ctypes.c_double)(get_cython_function_address('scipy.special.cython_special', '__pyx_fuse_1jv')),
    
    ('iv', numba.types.float64, numba.types.float64): ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double, ctypes.c_double)(get_cython_function_address('scipy.special.cython_special', '__pyx_fuse_1iv')),
    ('kv', numba.types.float64, numba.types.float64): ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double, ctypes.c_double)(get_cython_function_address('scipy.special.cython_special', '__pyx_fuse_1kv')),
    }

def _choose_kernel(name, all_signatures):
    def choice_function(*args):
        for signature in all_signatures:
            if args == signature:
                f = name_and_types_to_pointer[(name, *signature)]
                return lambda *args: f(*args)
    return choice_function

def _add_overloads():
    for name, all_signatures in name_to_numba_signatures.items():
        sc_function = getattr(scipy.special, name)
        numba.extending.overload(sc_function)(
            _choose_kernel(name, all_signatures)
        )

_add_overloads()



# =============================================================================
# helper to determine number of available CPU cores
# =============================================================================
def _get_nr_processes():
    """return available processes
    
    see: 
    https://stackoverflow.com/questions/1006289/how-to-find-out-the-number-of-cpus-using-python
    """
    ## preffered method to get available processes (might fail on windows)
    try:
        import os
        return len(os.sched_getaffinity(0))
    except:
        pass
    
    ## if failed, try alternative using psutils
    try:
        import psutil
        return len(psutil.Process().cpu_affinity())
    except:
        pass
    
    ## fall back on multiprocessing value (if psutils not installed)
    import multiprocessing
    return multiprocessing.cpu_count()



#==============================================================================
# globals
#==============================================================================
DTYPE_C = np.complex64

#************************************************************************
#************************************************************************
# Definition of HG, radially and azimuthally pol. doughnut modes following
# Novotny & Hecht Principle of Nano-Optics (p.62) extended to take into
# account an interface (transmission part) by G. Colas des Francs
# Direct calculus through integrals I_ii and K_ii of the azimuthally and 
# radially polarized doghnuts (+ numba for calculation acceleration)
#************************************************************************
#************************************************************************               

@numba.njit                    
def _fw(theta, f0, thetamax):

    """ Novotny & Hecht Principle of Nano-Optics - p62

    Parameters
    ----------
    theta : angle
    f0: lens filling factor (Novotny p62) f0 = w0/ (f sin theta max) 
    thetamax: max. angle allowed in lens
    Returns
    -------     
    
    Notes
    -----
         add notes
    
    """
    
    fres = np.exp(-(np.sin(theta)/f0/np.sin(thetamax))**2)
    return fres                  
                    
# Conversion cartersian to cylindrical
@numba.njit
def _cart_cyl(x,y):
    rho = np.sqrt(x**2 + y**2)
    phi = np.arctan2(y,x)
    return rho, phi   
   
# Wavevectors
@numba.jit(numba.typeof((1.0))(numba.double, numba.double, numba.double), nopython=True)                 
def _krho(n1,k0,theta):
    return n1*k0*np.sin(theta)

@numba.njit                    
def _kz1(n1,k0,theta):
    return n1*k0*np.cos(theta)

@numba.njit                    
def _kz3(n1,n3,k0,theta):
    return cmath.sqrt(n3**2 * k0**2 - _krho(n1,k0,theta)**2)

# Fresnel transmition and reflection coefficients for TE (s) and TM (p) polarization
@numba.njit                    
def _t_TE(n1,n3,k0,theta) :
    return 2.0 * _kz1(n1,k0,theta)/(_kz1(n1,k0,theta) + _kz3(n1,n3,k0,theta))

@numba.njit                    
def _t_TM(n1,n3,k0,theta) :
    return (2.0*n1*n3*_kz1(n1,k0,theta)/(n1**2 * _kz3(n1,n3,k0,theta) + n3**2 * _kz1(n1,k0,theta)))

@numba.njit                    
def _r_TE(n1,n3,k0,theta) :
    return ( _kz1(n1,k0,theta) - _kz3(n1,n3,3,k0,theta))/( _kz1(n1,k0,theta) + _kz3(n1,n3,k0,theta))

@numba.njit                    
def _r_TM(n1,n3,k0,theta) :
    return (n3**2 * _kz1(n1,k0,theta) - n1**2*_kz3(n1,k0,theta)) / (n3**2 * _kz1(n1,k0,theta) + n1**2 * _kz3(n1,n3,k0,theta))

# cos and sin theta in the outgoing media
@numba.njit
def _sintheta3(n1,n3,theta) :
    return n1/n3 * np.sin(theta)

@numba.njit
def _costheta3(n1,n3,theta) :
    ct3 = cmath.sqrt(1. - _sintheta3(n1, n3, theta)**2)
    ct3 *= np.sign(np.sign(ct3.imag) + 0.5)  # if imag<0: multiply by -1
    
    return ct3

"""
In all the following integrals (IXX_t and KXX_t) :
    Parameters
    ----------
    x,y,z : position
    k0: wavector in vaacum
    kSign: beam z-propagation direction
           -1: top to Bottom, 1 Bottom to top
    NA : numerical aperture
    n1: incident medium index
    n3: outgoing medium index 
    f : lens focal distance (mm)
    w0 : beam waist  (mm)
"""
                                     
# Electric part
@numba.njit
def _I00_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):    
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))
    rho, phi = _cart_cyl(x,y)
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)*(_t_TE(n1,n3,k0,theta) + _t_TM(n1,n3,k0,theta) * _costheta3(n1,n3,theta) )
              *scipy.special.jv(0.,kSign*_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))
    return result

def _I00_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _I00_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _I01_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)*_t_TM(n1,n3,k0,theta)*_sintheta3(n1,n3,theta)
              *scipy.special.jv(1.,_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))
    return result

def _I01_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _I01_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _I02_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)*(_t_TE(n1,n3,k0,theta) - _t_TM(n1,n3,k0,theta) * _costheta3(n1,n3,theta) )
              *scipy.special.jv(2.,_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z)) 
    return result

def _I02_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _I02_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _I10_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)**2*_t_TM(n1,n3,k0,theta)*_sintheta3(n1,n3,theta)
              *scipy.special.jv(0.,_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))
    return result

def _I10_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _I10_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _I11_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)**2*(_t_TE(n1,n3,k0,theta) + 3*_t_TM(n1,n3,k0,theta)*_costheta3(n1,n3,theta))
              *scipy.special.jv(1.,1.*_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))
    return result

def _I11_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _I11_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _I12_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)**2*(_t_TE(n1,n3,k0,theta) - _t_TM(n1,n3,k0,theta)*_costheta3(n1,n3,theta))
              *scipy.special.jv(1.,_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))
    return result

def _I12_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _I12_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _I13_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)**2*_t_TM(n1,n3,k0,theta)*_sintheta3(n1,n3,theta)
              *scipy.special.jv(2.,_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))
    return result

def _I13_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _I13_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _I14_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)**2*(_t_TE(n1,n3,k0,theta) - _t_TM(n1,n3,k0,theta)*_costheta3(n1,n3,theta))
              *scipy.special.jv(3.,_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))
    return result

def _I14_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _I14_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

# Magnetic part
@numba.njit
def _K00_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):        
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))
    rho, phi = _cart_cyl(x,y)
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)*(_t_TM(n1,n3,k0,theta) + _t_TE(n1,n3,k0,theta) * _costheta3(n1,n3,theta) )
              *scipy.special.jv(0.,kSign*_krho(n1,k0,theta)*rho)
              * np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))    
    return result

def _K00_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _K00_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _K01_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)*_t_TE(n1,n3,k0,theta)*_sintheta3(n1,n3,theta)
              *scipy.special.jv(1.,_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))   
    return result

def _K01_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _K01_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _K02_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
           *np.sin(theta)*(_t_TM(n1,n3,k0,theta) - _t_TE(n1,n3,k0,theta) * _costheta3(n1,n3,theta) )
           *scipy.special.jv(2.,_krho(n1,k0,theta)*rho)
           *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))
    return result

def _K02_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _K02_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _K10_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)**2*_t_TE(n1,n3,k0,theta)*_sintheta3(n1,n3,theta)
              *scipy.special.jv(0.,_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))
    return result

def _K10_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _K10_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _K11_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)**2*(3.*_t_TE(n1,n3,k0,theta)*_costheta3(n1,n3,theta) + _t_TM(n1,n3,k0,theta))
              *scipy.special.jv(1.,_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))    
    return result

def _K11_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _K11_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _K12_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (-1.*_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)**2*( _t_TE(n1,n3,k0,theta)*_costheta3(n1,n3,theta) - _t_TM(n1,n3,k0,theta))
              *scipy.special.jv(1., _krho(n1,k0,theta)*rho )
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))    
    return result

def _K12_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _K12_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _K13_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)**2*_t_TE(n1,n3,k0,theta)*_sintheta3(n1,n3,theta)
              *scipy.special.jv(2.,_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))
    return result

def _K13_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _K13_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

@numba.njit
def _K14_t_integral(theta,x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    f0 = w0/(f*np.sin(thetamax))    
    rho, phi = _cart_cyl(x,y)    
    result = (-1.*_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
              *np.sin(theta)**2*(_t_TE(n1,n3,k0,theta)*_costheta3(n1,n3,theta) - _t_TM(n1,n3,k0,theta))
              *scipy.special.jv(3.,_krho(n1,k0,theta)*rho)
              *np.exp(1j*n3*k0*_costheta3(n1,n3,theta)*kSign*z))   
    return result

def _K14_t(x,y,z,k0,kSign,n1,n3,NA,f,w0):
    thetamax = np.arcsin(NA/n1)
    int_func = lambda  theta: _K14_t_integral(theta, x,y,z,k0,kSign,n1,n3,NA,f,w0)
    integr, err = integrate.quad_vec(int_func, 0, thetamax, quadrature='gk15')
    return integr

def _func_calc_E_HG_00_t(args_list):
    x,y,z, k0,kSign,n1,n3,NA,f,w0,prefactor,polarization_state = args_list
    
    phi = np.arctan2(y,x)
 
    I_00  = _I00_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_02  = _I02_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_01  = _I01_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
 
    # Electric field
    # x-polarization
    Exx = kSign * (I_00 + I_02*np.cos(2*phi))
    Eyx = kSign * I_02 * np.sin(2*phi)
    Ezx = (-2 * 1j * I_01 * np.cos(phi) )    
    # y-polarization
    Exy = kSign * I_02 * np.sin(2*phi)
    Eyy = kSign * (I_00 - I_02*np.cos(2*phi))
    Ezy = (-2 * 1j * I_01 * np.sin(phi))
    
    ## amplitude and polarization
    E0x = polarization_state[0]
    E0y = polarization_state[1]
    Dphi = polarization_state[2]
    Aphi = polarization_state[3]
    
    # adding the polarization component
    Ex = (E0x * Exx  + E0y * Exy)
    Ey = (E0x * Eyx  + E0y * Eyy)
    Ez = (E0x * Ezx  + E0y * Ezy)
         
    # adding an absolute phase
    abs_phase = np.exp(1j * Aphi)
    Ex *= abs_phase
    Ey *= np.exp(1j * Dphi) * abs_phase    
      
    return prefactor * np.array([Ex, Ey, Ez])

def _func_calc_H_HG_00_t(args_list):
    x,y,z, k0,kSign,n1,n3,NA,f,w0,prefactor,polarization_state,Z0 = args_list
    
    phi = np.arctan2(y,x)
              
    K_00 = _K00_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_02 = _K02_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_01 = _K01_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    
    # Magnetic field
    # x-polarization
    Bxx = kSign * K_02 * np.sin(2*phi)
    Byx = kSign * (K_00 - K_02*np.cos(2*phi))
    Bzx = (-2 * 1j * K_01 * np.sin(phi))
    
    # y-polarization
    Bxy = kSign * (K_00 + K_02*np.cos(2*phi))
    Byy = kSign * K_02 * np.sin(2*phi)
    Bzy = (-2 * 1j * K_01 * np.cos(phi))        
    
    ## amplitude and polarization
    E0x = polarization_state[0]
    E0y = polarization_state[1]
    Dphi = polarization_state[2]
    Aphi = polarization_state[3]
    
    # adding the polarization component
    Bx = (E0x * Bxx  + E0y * Bxy)
    By = (E0x * Byx  + E0y * Byy)
    Bz = (E0x * Bzx  + E0y * Bzy)
         
    # adding an absolute phase
    abs_phase = np.exp(1j * Aphi)
    Bx *= abs_phase
    By *= np.exp(1j * Dphi) * abs_phase
    
    return prefactor * n3/Z0 * np.array([Bx, By, Bz])

def HermiteGauss00(
        pos, env_dict, wavelength, 
        theta=0, polarization_state=None,
        xSpot=0.0, ySpot=0.0, zSpot=0.0, kSign=-1.0,
        NA=0.5, f=100, w0=1, phase=0.0, normalize=False, returnField='E', N_cpu=-1):

    """Hermite Gaussian mode (0,0)
    
    after Novotny & Hecht p62 - Eq. (3.66)
    extended to transmission through an interface (G. Colas des Francs)
    
    
    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]
    
    env_dict : dict
        Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
        description of environment. Must contain ['eps1', 'eps2', 'eps3', 'spacing'].
    
    wavelength : float
          Wavelength in nm
      
    theta : float, optional
        linear polarization angle in degrees, 0deg = 0X direction.
        either 'theta' or 'polarization_state' must be given.
        The default is 0 (x polarization).
    
    polarization_state : 4-tuple of float, optional
        either 'theta' or 'polarization_state' must be given.
        polarization state with field amplitudes and phases, tuple of 4 float:
        (E0x, E0y, Dphi, Aphi): E0X amplitde, E0Y amplitde, phase difference 
        between X and Y components (in rad), absolute phase of plane wave (in rad).
        The field is then calculated as E = (E0x, E0y*exp(i*Dphi*z), 0)*exp(i*Aphi*z).
            Dphi : 
                - positive: left hand rotating polarization
                - negative: right hand rotating polarization 
                - example: left circular pol. with (E0x=1, E0y=1, Dphi=np.pi/2., phi=0)
        The default is `None`.
    
    xSpot, ySpot, zSpot : float, optional
        position of beam focus in nm. The default is 0,0,0
        
    kSign : float, optional
        Direction of Beam. -1: top to Bottom, 1 Bottom to top. 
        The default is -1
              
    NA : float, optional
        lens numerical aperture. The default is 0.5.
    
    f : float, optional
        lens focal distance (in mm). The default is 100.
    
    w0 : float, optional
        beam waist before focalisation (in mm). The default is 1.
    
    phase : float, optional
        additional phase (degrees). The default is 0.
    
    normalize : bool, optional
        Normalize field amplitude at focus. If False, field is 
        normalized *before* the focus. The default is False.
                     
    returnField  : str, optional
        'B' or 'E'. The default is 'E'.
    
    N_cpu : int, optional
        numbers of CPUs to use for beam evaluation. The default is -1 (all CPUs).
    """
    if N_cpu == -1:
        N_cpu = _get_nr_processes()
    
    if 'eps_env' in env_dict.keys():
        cn1 = cn2 = env_dict['eps_env']**0.5
    else:
        cn1 = env_dict['eps1']**0.5
        cn2 = env_dict['eps2']**0.5
        cn3 = env_dict['eps3']**0.5
        # spacing = env_dict['spacing']**0.5
#        if cn1 != cn2 or cn2 != cn3:
    if np.imag(cn1) != 0 or np.imag(cn2) != 0 or np.imag(cn3) != 0:
        warnings.warn("Special modes only support real refractive index")
    if cn2 != cn3:
        warnings.warn("Special modes only support one interface. Using ref.indices of medium 1 and 3, but results may be incorrect.")
    
    cn1 = np.real(cn1)
    cn2 = np.real(cn2)
    cn3 = np.real(cn3)
        
    f *= 1E6  # conversion mm > nm
    w0 *= 1E6  # conversion mm > nm
        
    xm, ym, zm = np.transpose(pos)    
    npts = len(xm)
 
    if kSign == -1 : 
        n1 = cn3
        n3 = cn1
    else:    
        n1 = cn1
        n3 = cn3
        
    if (theta is None and polarization_state is None) or (theta is not None and polarization_state is not None):
        raise ValueError("exactly one argument of 'theta' and 'polarization_state' must be given.")
    if theta is not None:
        polarization_state = (1.0 * np.cos(theta * np.pi/180.), 
                              1.0 * np.sin(theta * np.pi/180.), 
                              0, 0)
    #Vacuum impedence
    eps0 = 8.85418782e-12
    mu0  = 4e-7*np.pi
    Z0 = np.sqrt(mu0/eps0)    

    k = kSign * n1 * (2*np.pi / wavelength)
    k0 = 2*np.pi / wavelength
    prefactor = 0.5*1j*k*f*np.exp(-1j*k*f)
    
    #adding an additional phase to the beam, in degrees
    prefactor *= np.exp(1j*phase*np.pi/180.)
    
    if returnField.lower() == 'e':
        argslist = []
        for ipt in range(npts):
           
            x = xm[ipt] - xSpot
            y = ym[ipt] - ySpot
            z = zm[ipt] - zSpot
            
            argslist.append([x, y, z, k0, kSign, n1, n3, NA, f, w0,
                             prefactor, polarization_state])
     
        with multiprocessing.Pool(N_cpu) as p:
            E_list = p.map(_func_calc_E_HG_00_t, argslist)
        output = np.array(E_list, dtype=DTYPE_C)
    else:
        argslist = []
        for ipt in range(npts):
           
            x = xm[ipt] - xSpot
            y = ym[ipt] - ySpot
            z = zm[ipt] - zSpot

            argslist.append([x, y, z, k0, kSign, n1, n3, NA, f, w0,
                             prefactor, polarization_state, Z0])
            
        with multiprocessing.Pool(N_cpu) as p:
            B_list = p.map(_func_calc_H_HG_00_t, argslist)
        output = np.array(B_list, dtype=DTYPE_C)


    if normalize:
        return output / np.linalg.norm(output, axis=1).max()
    else:
        return output

def _func_calc_E_HG_10_t(args_list):
    x,y,z, k0,kSign,n1,n3,NA,f,w0,prefactor,polarization_state = args_list
    
    phi = np.arctan2(y,x)
 
    I_10 = _I10_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_11 = _I11_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_12 = _I12_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_13 = _I13_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_14 = _I14_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
 
    # Electric field
    # x-polarization
    Exx = kSign * 1j * ( I_11 * np.cos(phi) + I_14 * np.cos(3.*phi) )
    Eyx = kSign * 1j * ( -1.* I_12 * np.sin(phi) + I_14 * np.sin(3.*phi) )
    Ezx = 2.         * (-1. * I_10 + I_13 * np.cos(2.*phi) )    
    # y-polarization
    Exy = kSign * 1j * (-1. * I_12 * np.sin(phi) + I_14 * np.sin(3.*phi) )
    Eyy = kSign * 1j * ( (I_11 + 2.*I_12) * np.cos(phi) - I_14 * np.cos(3.*phi) )
    Ezy = 2.         * I_13 * np.sin(2.*phi) 
    
    ## amplitude and polarization
    E0x = polarization_state[0]
    E0y = polarization_state[1]
    Dphi = polarization_state[2]
    Aphi = polarization_state[3]
    
    # adding the polarization component
    Ex = (E0x * Exx  + E0y * Exy)
    Ey = (E0x * Eyx  + E0y * Eyy)
    Ez = (E0x * Ezx  + E0y * Ezy)
         
    # adding an absolute phase
    abs_phase = np.exp(1j * Aphi)
    Ex *= abs_phase
    Ey *= np.exp(1j * Dphi) * abs_phase    
      
    return prefactor * np.array([Ex, Ey, Ez])

def _func_calc_H_HG_10_t(args_list):
    x,y,z, k0,kSign,n1,n3,NA,f,w0,prefactor,polarization_state,Z0 = args_list
    
    phi = np.arctan2(y,x)
    
    K_10 = _K10_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_11 = _K11_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_12 = _K12_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_13 = _K13_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_14 = _K14_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
            
    # Magnetic field
    # x-polarization
    Bxx = kSign * 1j * ( -1.* K_12 * np.sin(phi) + K_14 * np.sin(3.*phi) )
    Byx = kSign * 1j * ( (K_11 + 2.*K_12)*np.cos(phi) - K_14 * np.cos(3.*phi) )
    Bzx = 2.         * K_13  * np.sin(2.* phi)
         
    # y-polarization
    Bxy = kSign * 1j * ( -1.* K_11 * np.cos(phi) - K_14 * np.cos(3.*phi) )
    Byy = kSign * 1j * ( K_12 * np.sin(phi) - K_14 * np.sin(3.*phi) )
    Bzy = 2.         * ( K_10 - K_13 * np.cos(2.*phi) )
    
    ## amplitude and polarization
    E0x = polarization_state[0]
    E0y = polarization_state[1]
    Dphi = polarization_state[2]
    Aphi = polarization_state[3]
    
    # adding the polarization component
    Bx = (E0x * Bxx  + E0y * Bxy)
    By = (E0x * Byx  + E0y * Byy)
    Bz = (E0x * Bzx  + E0y * Bzy)
         
    # adding an absolute phase
    abs_phase = np.exp(1j * Aphi)
    Bx *= abs_phase
    By *= np.exp(1j * Dphi) * abs_phase
    
    return prefactor * n3/Z0 * np.array([Bx, By, Bz])

def HermiteGauss10(
        pos, env_dict, wavelength, 
        theta=0, polarization_state=None,
        xSpot=0.0, ySpot=0.0, zSpot=0.0, kSign=-1.0,
        NA=0.5, f=100, w0=1, phase=0.0, normalize=False, returnField='E', N_cpu=-1):

    """Hermite Gaussian mode (1,0)
    
    after Novotny & Hecht p62 - Eq. (3.66)
    extended to transmission through an interface (G. Colas des Francs)
    
    
    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]
    
    env_dict : dict
        Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
        description of environment. Must contain ['eps1', 'eps2', 'eps3', 'spacing'].
    
    wavelength : float
          Wavelength in nm
      
    theta : float, optional
        linear polarization angle in degrees, 0deg = 0X direction.
        either 'theta' or 'polarization_state' must be given.
        The default is 0 (x polarization).
    
    polarization_state : 4-tuple of float, optional
        either 'theta' or 'polarization_state' must be given.
        polarization state with field amplitudes and phases, tuple of 4 float:
        (E0x, E0y, Dphi, Aphi): E0X amplitde, E0Y amplitde, phase difference 
        between X and Y components (in rad), absolute phase of plane wave (in rad).
        The field is then calculated as E = (E0x, E0y*exp(i*Dphi*z), 0)*exp(i*Aphi*z).
            Dphi : 
                - positive: left hand rotating polarization
                - negative: right hand rotating polarization 
                - example: left circular pol. with (E0x=1, E0y=1, Dphi=np.pi/2., phi=0)
        The default is `None`.
    
    xSpot, ySpot, zSpot : float, optional
        position of beam focus in nm. The default is 0,0,0
        
    kSign : float, optional
        Direction of Beam. -1: top to Bottom, 1 Bottom to top. 
        The default is -1
              
    NA : float, optional
        lens numerical aperture. The default is 0.5.
    
    f : float, optional
        lens focal distance (in mm). The default is 100.
    
    w0 : float, optional
        beam waist before focalisation (in mm). The default is 1.
    
    phase : float, optional
        additional phase (degrees). The default is 0.
    
    normalize : bool, optional
        Normalize field amplitude at focus. If False, field is 
        normalized *before* the focus. The default is False.
                                  
    returnField  : str, optional
        'B' or 'E'. The default is 'E'.
    
    N_cpu : int, optional
        numbers of CPUs to use for beam evaluation. The default is -1 (all CPUs).
    """
    if N_cpu == -1:
        N_cpu = _get_nr_processes()
    
    if 'eps_env' in env_dict.keys():
        cn1 = cn2 = env_dict['eps_env']**0.5
    else:
        cn1 = env_dict['eps1']**0.5
        cn2 = env_dict['eps2']**0.5
        cn3 = env_dict['eps3']**0.5
        # spacing = env_dict['spacing']**0.5
#        if cn1 != cn2 or cn2 != cn3:
    if np.imag(cn1) != 0 or np.imag(cn2) != 0 or np.imag(cn3) != 0:
        warnings.warn("Special modes only support real refractive index")
    if cn2 != cn3:
        warnings.warn("Special modes only support one interface. Using ref.indices of medium 1 and 3, but results may be incorrect.")
    
    cn1 = np.real(cn1)
    cn2 = np.real(cn2)
    cn3 = np.real(cn3)
        
    f *= 1E6  # conversion mm > nm
    w0 *= 1E6  # conversion mm > nm
        
    xm, ym, zm = np.transpose(pos)    
    npts = len(xm)
 
    if kSign == -1 : 
        n1 = cn3
        n3 = cn1
    else:    
        n1 = cn1
        n3 = cn3
        
    if (theta is None and polarization_state is None) or (theta is not None and polarization_state is not None):
        raise ValueError("exactly one argument of 'theta' and 'polarization_state' must be given.")
    if theta is not None:
        polarization_state = (1.0 * np.cos(theta * np.pi/180.), 
                              1.0 * np.sin(theta * np.pi/180.), 
                              0, 0)
    #Vacuum impedence
    eps0 = 8.85418782e-12
    mu0  = 4e-7*np.pi
    Z0 = np.sqrt(mu0/eps0)    

    k = kSign * n1 * (2*np.pi / wavelength)
    k0 = 2*np.pi / wavelength
    prefactor = 0.5/w0*1j*k*f**2*np.exp(-1j*k*f)
    
    #adding a additional phase to the beam, in degrees
    prefactor *= np.exp(1j*phase*np.pi/180.)
    
    if returnField.lower() == 'e':
        argslist = []
        for ipt in range(npts):
           
            x = xm[ipt] - xSpot
            y = ym[ipt] - ySpot
            z = zm[ipt] - zSpot

            argslist.append([x, y, z, k0, kSign, n1, n3, NA, f, w0,
                             prefactor, polarization_state])
              
        with multiprocessing.Pool(N_cpu) as p:
            E_list = p.map(_func_calc_E_HG_10_t, argslist)
        output = np.array(E_list, dtype=DTYPE_C)

    else :
        argslist = []
        for ipt in range(npts):
           
            x = xm[ipt] - xSpot
            y = ym[ipt] - ySpot
            z = zm[ipt] - zSpot
         
            argslist.append([x, y, z, k0, kSign, n1, n3, NA, f, w0,
                             prefactor, polarization_state, Z0])
            
        with multiprocessing.Pool(N_cpu) as p:
            B_list = p.map(_func_calc_H_HG_10_t, argslist)
        output = np.array(B_list, dtype=DTYPE_C)

    if normalize:
        return output / np.linalg.norm(output, axis=1).max()
    else:
        return output

def _func_calc_E_HG_01_t(args_list):
    x,y,z, k0,kSign,n1,n3,NA,f,w0,prefactor,polarization_state = args_list
    
    phi = np.arctan2(y,x)
 
    I_10 = _I10_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_11 = _I11_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_12 = _I12_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_13 = _I13_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_14 = _I14_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
 
    # Electric field
    # x-polarization
    Exx = kSign * 1j * ( (I_11 + 2.*I_12) * np.sin(phi) + I_14 * np.sin(3.*phi) )
    Eyx = kSign *-1j * ( I_12 * np.cos(phi) + I_14 * np.cos(3.*phi) )
    Ezx = 2.         *   I_13 * np.sin(2.*phi)    
    # y-polarization
    Exy = kSign *-1j * ( I_12 * np.cos(phi) + I_14 * np.cos(3.*phi) )
    Eyy = kSign * 1j * ( I_11 * np.sin(phi) - I_14 * np.sin(3.*phi) )
    Ezy = -2.        * (I_10 + I_13 * np.cos(2.*phi) )
    
    ## amplitude and polarization
    E0x = polarization_state[0]
    E0y = polarization_state[1]
    Dphi = polarization_state[2]
    Aphi = polarization_state[3]
    
    # adding the polarization component
    Ex = (E0x * Exx  + E0y * Exy)
    Ey = (E0x * Eyx  + E0y * Eyy)
    Ez = (E0x * Ezx  + E0y * Ezy)
         
    # adding an absolute phase
    abs_phase = np.exp(1j * Aphi)
    Ex *= abs_phase
    Ey *= np.exp(1j * Dphi) * abs_phase   
    
    return prefactor * np.array([Ex, Ey, Ez])

def _func_calc_H_HG_01_t(args_list):
    x,y,z, k0,kSign,n1,n3,NA,f,w0,prefactor,polarization_state,Z0 = args_list
    phi = np.arctan2(y,x)
    
    K_10 = _K10_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_11 = _K11_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_12 = _K12_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_13 = _K13_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_14 = _K14_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
            
    # Magnetic field
    # x-polarization
    Bxx = kSign * 1j * (-K_12 * np.cos(phi) - K_14 * np.cos(3.*phi) )
    Byx = kSign * 1j * ( K_11 * np.sin(phi) - K_14 * np.sin(3.*phi) )
    Bzx = 2.         * (-K_10 - K_13  * np.cos(2.* phi) ) 
         
    # y-polarization
    Bxy = kSign * 1j * (-(K_11 + 2.*K_12) * np.sin(phi) - K_14 * np.sin(3.*phi) )
    Byy = kSign * 1j * ( K_12 * np.cos(phi) + K_14 * np.cos(3.*phi) )
    Bzy = -2.         *  K_13 * np.sin(2.*phi)
    
    ## amplitude and polarization
    E0x = polarization_state[0]
    E0y = polarization_state[1]
    Dphi = polarization_state[2]
    Aphi = polarization_state[3]
    
    # adding the polarization component
    Bx = (E0x * Bxx  + E0y * Bxy)
    By = (E0x * Byx  + E0y * Byy)
    Bz = (E0x * Bzx  + E0y * Bzy)
         
    # adding an absolute phase
    abs_phase = np.exp(1j * Aphi)
    Bx *= abs_phase
    By *= np.exp(1j * Dphi) * abs_phase
    
    return prefactor * n3/Z0 * np.array([Bx, By, Bz])

def HermiteGauss01(
        pos, env_dict, wavelength, 
        theta=0, polarization_state=None,
        xSpot=0.0, ySpot=0.0, zSpot=0.0, kSign=-1.0,
        NA=0.5, f=100, w0=1, phase=0.0, normalize=False, returnField='E', N_cpu=-1):
    
    """Gaussian mode (0,1)
    
    after Novotny & Hecht p62 - Eq. (3.66)
    extended to transmission through an interface (G. Colas des Francs)
    
    
    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]
    
    env_dict : dict
        Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
        description of environment. Must contain ['eps1', 'eps2', 'eps3', 'spacing'].
    
    wavelength : float
          Wavelength in nm
      
    theta : float, optional
        linear polarization angle in degrees, 0deg = 0X direction.
        either 'theta' or 'polarization_state' must be given.
        The default is 0 (x polarization).
    
    polarization_state : 4-tuple of float, optional
        either 'theta' or 'polarization_state' must be given.
        polarization state with field amplitudes and phases, tuple of 4 float:
        (E0x, E0y, Dphi, Aphi): E0X amplitde, E0Y amplitde, phase difference 
        between X and Y components (in rad), absolute phase of plane wave (in rad).
        The field is then calculated as E = (E0x, E0y*exp(i*Dphi*z), 0)*exp(i*Aphi*z).
            Dphi : 
                - positive: left hand rotating polarization
                - negative: right hand rotating polarization 
                - example: left circular pol. with (E0x=1, E0y=1, Dphi=np.pi/2., phi=0)
        The default is `None`.
    
    xSpot, ySpot, zSpot : float, optional
        position of beam focus in nm. The default is 0,0,0
        
    kSign : float, optional
        Direction of Beam. -1: top to Bottom, 1 Bottom to top. 
        The default is -1
              
    NA : float, optional
        lens numerical aperture. The default is 0.5.
    
    f : float, optional
        lens focal distance (in mm). The default is 100.
    
    w0 : float, optional
        beam waist before focalisation (in mm). The default is 1.
    
    phase : float, optional
        additional phase (degrees). The default is 0.
    
    normalize : bool, optional
        Normalize field amplitude at focus. If False, field is 
        normalized *before* the focus. The default is False.
    
    returnField  : str, optional
        'B' or 'E'. The default is 'E'.     
    
    N_cpu : int, optional
        numbers of CPUs to use for beam evaluation. The default is -1 (all CPUs).
    """
    if N_cpu == -1:
        N_cpu = _get_nr_processes()
  
    
    if 'eps_env' in env_dict.keys():
        cn1 = cn2 = env_dict['eps_env']**0.5
    else:
        cn1 = env_dict['eps1']**0.5
        cn2 = env_dict['eps2']**0.5
        cn3 = env_dict['eps3']**0.5
        # spacing = env_dict['spacing']**0.5
#        if cn1 != cn2 or cn2 != cn3:
    if np.imag(cn1) != 0 or np.imag(cn2) != 0 or np.imag(cn3) != 0:
        warnings.warn("Special modes only support real refractive index")
    if cn2 != cn3:
        warnings.warn("Special modes only support one interface. Using ref.indices of medium 1 and 3, but results may be incorrect.")
    
    cn1 = np.real(cn1)
    cn2 = np.real(cn2)
    cn3 = np.real(cn3)
        
    f *= 1E6  # conversion mm > nm
    w0 *= 1E6  # conversion mm > nm
        
    xm, ym, zm = np.transpose(pos)    
    npts = len(xm)
 
    if kSign == -1 : 
        n1 = cn3
        n3 = cn1
    else:    
        n1 = cn1
        n3 = cn3
        
    if (theta is None and polarization_state is None) or (theta is not None and polarization_state is not None):
        raise ValueError("exactly one argument of 'theta' and 'polarization_state' must be given.")
    if theta is not None:
        polarization_state = (1.0 * np.cos(theta * np.pi/180.), 
                              1.0 * np.sin(theta * np.pi/180.), 
                              0, 0)
    
    #Vacuum impedence
    eps0 = 8.85418782e-12
    mu0  = 4e-7*np.pi
    Z0 = np.sqrt(mu0/eps0)    

    k = kSign * n1 * (2*np.pi / wavelength)
    k0 = 2*np.pi / wavelength
    prefactor = 0.5/w0*1j*k*f**2*np.exp(-1j*k*f)
    
    #adding a additional phase to the beam, in degrees
    prefactor *= np.exp(1j*phase*np.pi/180.)
    
    if returnField.lower() == 'e':
        argslist = []        
        for ipt in range(npts):
           
            x = xm[ipt] - xSpot
            y = ym[ipt] - ySpot
            z = zm[ipt] - zSpot
    
            argslist.append([x, y, z, k0, kSign, n1, n3, NA, f, w0,
                             prefactor,polarization_state]) 
                        
        with multiprocessing.Pool(N_cpu) as p:
            E_list = p.map(_func_calc_E_HG_01_t, argslist)
        output = np.array(E_list, dtype=DTYPE_C)
            
    else :
        argslist = []
        for ipt in range(npts):
           
            x = xm[ipt] - xSpot
            y = ym[ipt] - ySpot
            z = zm[ipt] - zSpot
    
            argslist.append([x, y, z, k0, kSign, n1, n3, NA, f, w0,
                             prefactor,polarization_state, Z0])
            
        with multiprocessing.Pool(N_cpu) as p:
            B_list = p.map(_func_calc_H_HG_01_t, argslist)
        output = np.array(B_list, dtype=DTYPE_C)


    if normalize:
        return output / np.linalg.norm(output, axis=1).max()
    else:
        return output


def _func_calc_E_doughnut_rad_t(args_list):
    x,y,z, k0,kSign,n1,n3,NA,f,w0,prefactor,polarization_direction = args_list
    
    phi = np.arctan2(y,x)
 
    I_10 = _I10_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_11 = _I11_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_12 = _I12_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
 
    # Electric field
    Exi = kSign * 1j * ( (I_11 - I_12) * np.cos(phi) )
    Eyi = kSign * 1j * ( (I_11 - I_12) * np.sin(phi) )
    Ezi = -4.         *   I_10    
    
    return prefactor * polarization_direction * np.array([Exi, Eyi, Ezi])

def _func_calc_H_doughnut_rad_t(args_list):
    x,y,z, k0,kSign,n1,n3,NA,f,w0,prefactor,polarization_direction,Z0 = args_list
    
    phi = np.arctan2(y,x)
    
    K_11 = _K11_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_12 = _K12_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
            
    # Magnetic field
    Bxi = kSign * -1j * ( (K_11 + 3*K_12)*np.sin(phi) )
    Byi = kSign *  1j * ( (K_11 + 3*K_12)*np.cos(phi) )
    Bzi = 0.         

    return prefactor * n3/Z0 * polarization_direction * np.array([Bxi, Byi, Bzi])

def Radial_pol_doughnut(
        pos, env_dict, wavelength, 
        polarization_direction = 1.0, 
        xSpot=0.0, ySpot=0.0, zSpot=0.0, kSign=-1.0,
        NA=0.5, f=100, w0=1, phase=0.0, normalize=False, returnField='E', N_cpu=-1):
    """Radially polarized doughnut mode focused beam
    
    after Novotny & Hecht p67 - Eq. (3.70)
    extended to transmission through an interface (G. Colas des Francs)
    
    
    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]
    
    env_dict : dict
        Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
        description of environment. Must contain ['eps1', 'eps2', 'eps3', 'spacing'].
    
    wavelength : float
          Wavelength in nm
    
    polarization_direction : float, optional
        Either 1 or -1. Direction of the azimuthal polarization. 
        Equivalent to a phase of pi. The default is 1.
    
    xSpot, ySpot, zSpot : float, optional
        position of beam focus in nm. The default is 0,0,0
        
    kSign : float, optional
        Direction of Beam. -1: top to Bottom, 1 Bottom to top. 
        The default is -1
              
    NA : float, optional
        lens numerical aperture. The default is 0.5.
    
    f : float, optional
        lens focal distance (in mm). The default is 100.
    
    w0 : float, optional
        beam waist before focalisation (in mm). The default is 1.
    
    phase : float, optional
        additional phase (degrees). The default is 0.
                     
    returnField  : str, optional
        'B' or 'E'. The default is 'E'.    
    
    normalize : bool, optional
        Normalize field amplitude at focus. If False, field is 
        normalized *before* the focus. The default is False.
    
    N_cpu : int, optional
        numbers of CPUs to use for beam evaluation. The default is -1 (all CPUs).
    """
    if N_cpu == -1:
        N_cpu = _get_nr_processes()
    
    if 'eps_env' in env_dict.keys():
        cn1 = cn2 = env_dict['eps_env']**0.5
    else:
        cn1 = env_dict['eps1']**0.5
        cn2 = env_dict['eps2']**0.5
        cn3 = env_dict['eps3']**0.5
        # spacing = env_dict['spacing']**0.5
#        if cn1 != cn2 or cn2 != cn3:
    if np.imag(cn1) != 0 or np.imag(cn2) != 0 or np.imag(cn3) != 0:
        warnings.warn("Special modes only support real refractive index")
    if cn2 != cn3:
        warnings.warn("Special modes only support one interface. Using ref.indices of medium 1 and 3, but results may be incorrect.")
    
    cn1 = np.real(cn1)
    cn2 = np.real(cn2)
    cn3 = np.real(cn3)
        
    f *= 1E6  # conversion mm > nm
    w0 *= 1E6  # conversion mm > nm
        
    xm, ym, zm = np.transpose(pos)    
    npts = len(xm)
 
    if kSign == -1 : 
        n1 = cn3
        n3 = cn1
    else:    
        n1 = cn1
        n3 = cn3
        
    #Vacuum impedence
    eps0 = 8.85418782e-12
    mu0  = 4e-7*np.pi
    Z0 = np.sqrt(mu0/eps0)    

    k = kSign * n1 * (2*np.pi / wavelength)
    k0 = 2*np.pi / wavelength
    prefactor = 0.5/w0*1j*k*f**2*np.exp(-1j*k*f)
    
    #adding a additional phase to the beam, in degrees
    prefactor *= np.exp(1j*phase*np.pi/180.)
    
    if returnField.lower() == 'e':
        argslist = []
        for ipt in range(npts):
            x = xm[ipt] - xSpot
            y = ym[ipt] - ySpot
            z = zm[ipt] - zSpot
            argslist.append([x, y, z, k0, kSign, n1, n3, NA, f, w0,
                             prefactor, polarization_direction])
        
        with multiprocessing.Pool(N_cpu) as p:
            E_list = p.map(_func_calc_E_doughnut_rad_t, argslist)
        output = np.array(E_list, dtype=DTYPE_C)
    else :
        argslist = []
        for ipt in range(npts):
           
            x = xm[ipt] - xSpot
            y = ym[ipt] - ySpot
            z = zm[ipt] - zSpot
            argslist.append([x, y, z, k0, kSign, n1, n3, NA, f, w0,
                             prefactor, polarization_direction, Z0])
            
        with multiprocessing.Pool(N_cpu) as p:
            B_list = p.map(_func_calc_H_doughnut_rad_t, argslist)
        output = np.array(B_list, dtype=DTYPE_C)
        
    if normalize:
        return output / np.linalg.norm(output, axis=1).max()
    else:
        return output

def _func_calc_E_doughnut_azim_t(args_list):
    x,y,z, k0,kSign,n1,n3,NA,f,w0,prefactor,polarization_direction = args_list
    
    phi = np.arctan2(y,x)
 
    I_11 = _I11_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    I_12 = _I12_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
 
    # Electric field
    Exi = kSign * -1j * ( (I_11 + 3*I_12) * np.sin(phi) )
    Eyi = kSign *  1j * ( (I_11 + 3*I_12) * np.cos(phi) )
    Ezi= 0.
        
    return prefactor * polarization_direction * np.array([Exi, Eyi, Ezi])

def _func_calc_H_doughnut_azim_t(args_list):
    x,y,z, k0,kSign,n1,n3,NA,f,w0,prefactor,polarization_direction,Z0 = args_list
    
    phi = np.arctan2(y,x)
    
    K_10 = _K10_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_11 = _K11_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
    K_12 = _K12_t(x,y,z,k0,kSign,n1,n3,NA,f,w0)
            
    # Magnetic field
    Bxi = kSign * 1j * ( (K_12 - K_11) * np.cos(phi) )
    Byi = kSign * 1j * ( (K_12 - K_11) * np.sin(phi) )
    Bzi = kSign *  4 *    K_10    

    return prefactor * polarization_direction * n3/Z0 * np.array([Bxi, Byi, Bzi])


def Azimuth_pol_doughnut(
        pos, env_dict, wavelength, polarization_direction=1.0,
        xSpot=0.0, ySpot=0.0, zSpot=0.0, kSign=-1.0,
        NA=0.5, f=100, w0=1, phase=0.0, normalize=False, returnField='E', N_cpu=-1):

    """Azimuthally polarized doughnut
    
    After Novotny & Hecht p.67 - Eq. (3.70)
    Extended to include transmission through an interface (G. Colas des Francs)
    
    Parameters
    ----------
    pos : np.array
        list of 3-tuple coordinates to evaluate field at:
            [[x1,y1,z1], [x2,y2,z2], ... ]
    
    env_dict : dict
        Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
        description of environment. Must contain ['eps1', 'eps2', 'eps3', 'spacing'].
    
    wavelength : float
          Wavelength in nm
    
    polarization_direction : float, optional
        Either 1 or -1. Direction of the azimuthal polarization. 
        Equivalent to a phase of pi. The default is 1.
    
    xSpot, ySpot, zSpot : float, optional
        position of beam focus in nm. The default is 0,0,0
        
    kSign : float, optional
        Direction of Beam. -1: top to Bottom, 1 Bottom to top. 
        The default is -1
              
    NA : float, optional
        lens numerical aperture. The default is 0.5.
    
    f : float, optional
        lens focal distance (in mm). The default is 100.
    
    w0 : float, optional
        beam waist before focalisation (in mm). The default is 1.
    
    phase : float, optional
        additional phase (degrees). The default is 0.
    
    normalize : bool, optional
        Normalize field amplitude at focus. If False, field is 
        normalized *before* the focus. The default is False.
    
    returnField  : str, optional
        'B' or 'E'. The default is 'E'.
        
    N_cpu : int, optional
        numbers of CPUs to use for beam evaluation. The default is -1 (all CPUs).
    """
    if N_cpu == -1:
        N_cpu = _get_nr_processes()

    if 'eps_env' in env_dict.keys():
        cn1 = cn2 = env_dict['eps_env']**0.5
    else:
        cn1 = env_dict['eps1']**0.5
        cn2 = env_dict['eps2']**0.5
        cn3 = env_dict['eps3']**0.5
        # spacing = env_dict['spacing']**0.5
#        if cn1 != cn2 or cn2 != cn3:
    if np.imag(cn1) != 0 or np.imag(cn2) != 0 or np.imag(cn3) != 0:
        warnings.warn("Special modes only support real refractive index")
    if cn2 != cn3:
        warnings.warn("Special modes only support one interface. Using ref.indices of medium 1 and 3, but results may be incorrect.")
    
    cn1 = np.real(cn1)
    cn2 = np.real(cn2)
    cn3 = np.real(cn3)
        
    f *= 1E6  # conversion mm > nm
    w0 *= 1E6  # conversion mm > nm
        
    xm, ym, zm = np.transpose(pos)    
    npts = len(xm)
 
    if kSign == -1 : 
        n1 = cn3
        n3 = cn1
    else:    
        n1 = cn1
        n3 = cn3
        
    #Vacuum impedence
    eps0 = 8.85418782e-12
    mu0  = 4e-7*np.pi
    Z0 = np.sqrt(mu0/eps0)    

    k = kSign * n1 * (2*np.pi / wavelength)
    k0 = 2*np.pi / wavelength
    prefactor = 0.5/w0*1j*k*f**2*np.exp(-1j*k*f)
    
    #adding a additional phase to the beam, in degrees
    prefactor *= np.exp(1j*phase*np.pi/180.)
    
    if returnField.lower() == 'e':
        argslist = []
        for ipt in range(npts):
            x = xm[ipt] - xSpot
            y = ym[ipt] - ySpot
            z = zm[ipt] - zSpot
            argslist.append([x, y, z, k0, kSign, n1, n3, NA, f, w0,
                             prefactor, polarization_direction])
        with multiprocessing.Pool(N_cpu) as p:
            E_list = p.map(_func_calc_E_doughnut_azim_t, argslist)    
        output = np.array(E_list, dtype=DTYPE_C)
    else :
        argslist = []
        for ipt in range(npts):   
            x = xm[ipt] - xSpot
            y = ym[ipt] - ySpot
            z = zm[ipt] - zSpot
            argslist.append([x, y, z, k0, kSign, n1, n3, NA, f, w0,
                             prefactor, polarization_direction, Z0])
            
        with multiprocessing.Pool(N_cpu) as p:
            B_list = p.map(_func_calc_H_doughnut_azim_t, argslist)
        output = np.array(B_list, dtype=DTYPE_C)

     
    if normalize:
        return output / np.linalg.norm(output, axis=1).max()
    else:
        return output


if __name__ == "__main__":
    pass


# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
    Package providing different field generators 
    as illumination sources for pyGDM2
    
    - fast electron fields by A. Arbouet
    - focused vectorbeams by A. Arbouet, Y. Brûlé and G. Colas des Francs
    
"""
__name__ = 'pyGDM2.fields'
__author__ = 'Peter R. Wiecha'



## populate namespace
from pyGDM2.fields import efield_class
from pyGDM2.fields import regular
from pyGDM2.fields import focused_beams
from pyGDM2.fields import electron
from pyGDM2.fields import deprecated

from pyGDM2.fields.efield_class import efield

from pyGDM2.fields.regular import nullfield
from pyGDM2.fields.regular import plane_wave, gaussian
from pyGDM2.fields.regular import dipole_electric, dipole_magnetic

from pyGDM2.fields.focused_beams import HermiteGauss00, HermiteGauss01, HermiteGauss10
from pyGDM2.fields.focused_beams import Radial_pol_doughnut, Azimuth_pol_doughnut

from pyGDM2.fields.electron import fast_electron

from pyGDM2.fields.deprecated import planewave, focused_planewave, evanescent_planewave


## for pygdmUI graphical user interface:
FIELDS_LIST = [plane_wave, gaussian, 
               dipole_electric, dipole_magnetic,
               focused_planewave, 
               HermiteGauss00, HermiteGauss01, HermiteGauss10,
               Radial_pol_doughnut, Azimuth_pol_doughnut,
               fast_electron]


#!/usr/bin/env python
# -*- coding: utf-8 -*-

## --- for benchmark: limit openmp to 1 thread (for parallel scipy/numpy routines)
from __future__ import print_function, division

import os
nthreads = 1
os.environ["MKL_NUM_THREADS"] = "{}".format(int(nthreads))
os.environ["NUMEXPR_NUM_THREADS"] = "{}".format(int(nthreads))
os.environ["OMP_NUM_THREADS"] = "{}".format(int(nthreads))


from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import fields

from pyGDM2 import core

import numpy as np


## --- Note: It is not necessary to load mpi4py within the simulation script, this 
## --- will be done automatically by pyGDM2 prior to the actual MPI-simulation. 
## --- We do it however at this point to do some output to stdout only from 
## --- the master process (rank == 0).
from mpi4py import MPI
rank = MPI.COMM_WORLD.rank



#==============================================================================
# Configure simulation
#==============================================================================
## ---------- Setup structure
mesh = 'cube'
step = 20.0
radius = 3.5
geometry = structures.sphere(step, R=radius, mesh=mesh)
material = materials.dummy(2.0)
n1, n2 = 1.0, 1.0
struct = structures.struct(step, geometry, material, n1,n2, 
                           structures.get_normalization(mesh))

## ---------- Setup incident field
field_generator = fields.planewave
wavelengths = np.linspace(400, 800, 20)
kwargs = dict(theta = [0.0])
efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)

## ---------- Simulation initialization
sim = core.simulation(struct, efield)



#==============================================================================
# Rum the simulation
#==============================================================================
## --- mpi: print in process with rank=0, to avoid flooding of stdout
if rank == 0: 
    print("performing MPI parallel simulation... ")

core.scatter_mpi(sim)

if rank == 0: 
    print("simulation done.")

